from collections.abc import Callable
from dataclasses import dataclass
from typing import Any, Final, final

from soia._impl.function_maker import Expr, ExprLike
from soia._impl.timestamp import Timestamp
from soia._impl.type_adapter import TypeAdapter

from soia import _spec, reflection


class AbstractPrimitiveAdapter(TypeAdapter):
    @final
    def finalize(
        self,
        resolve_type_fn: Callable[[_spec.Type], "TypeAdapter"],
    ) -> None:
        pass

    @final
    def register_records(
        self,
        registry: dict[str, reflection.Record],
    ) -> None:
        pass


class _BoolAdapter(AbstractPrimitiveAdapter):
    def default_expr(self) -> ExprLike:
        return "False"

    def to_frozen_expr(self, arg_expr: ExprLike) -> Expr:
        return Expr.join("(True if ", arg_expr, " else False)")

    def is_not_default_expr(self, arg_expr: ExprLike, attr_expr: ExprLike) -> ExprLike:
        return arg_expr

    def to_json_expr(
        self,
        in_expr: ExprLike,
        readable: bool,
    ) -> ExprLike:
        if readable:
            return Expr.join("(True if ", in_expr, " else False)")
        else:
            return Expr.join("(1 if ", in_expr, " else 0)")

    def from_json_expr(self, json_expr: ExprLike) -> Expr:
        return Expr.join("(True if ", json_expr, " else False)")

    def get_type(self) -> reflection.Type:
        return reflection.PrimitiveType(
            kind="primitive",
            value="bool",
        )


BOOL_ADAPTER: Final[TypeAdapter] = _BoolAdapter()


@dataclass(frozen=True)
class _AbstractIntAdapter(AbstractPrimitiveAdapter):
    """Type adapter implementation for int32, int64 and uint64."""

    def default_expr(self) -> ExprLike:
        return "0"

    def to_frozen_expr(self, arg_expr: ExprLike) -> Expr:
        # Must accept float inputs and turn them into ints.
        return Expr.join("(0).__class__(", arg_expr, ")")

    def is_not_default_expr(self, arg_expr: ExprLike, attr_expr: ExprLike) -> ExprLike:
        return arg_expr

    def from_json_expr(self, json_expr: ExprLike) -> Expr:
        # Must accept float inputs and string inputs and turn them into ints.
        return Expr.join(
            "(0).__class__(",
            json_expr,
            ")",
        )


@dataclass(frozen=True)
class _Int32Adapter(_AbstractIntAdapter):
    def to_json_expr(self, in_expr: ExprLike, readable: bool) -> Expr:
        return Expr.join(
            "(-2147483648 if ",
            in_expr,
            " <= -2147483648 else ",
            in_expr,
            " if ",
            in_expr,
            " < 2147483647 else 2147483647)",
        )

    def get_type(self) -> reflection.Type:
        return reflection.PrimitiveType(
            kind="primitive",
            value="int32",
        )


def _int64_to_json(i: int) -> int | str:
    if i < -9007199254740991:  # min safe integer in JavaScript
        if i <= -9223372036854775808:
            return "-9223372036854775808"
        else:
            return str(i)
    elif i <= 9007199254740991:  # max safe integer in JavaScript
        return i
    elif i < 9223372036854775807:
        return str(i)
    else:
        return "9223372036854775807"


@dataclass(frozen=True)
class _Int64Adapter(_AbstractIntAdapter):
    def to_json_expr(self, in_expr: ExprLike, readable: bool) -> Expr:
        return Expr.join(Expr.local("int64_to_json", _int64_to_json), "(", in_expr, ")")

    def get_type(self) -> reflection.Type:
        return reflection.PrimitiveType(
            kind="primitive",
            value="int64",
        )


def _uint64_to_json(i: int) -> int | str:
    if i <= 0:
        return 0
    elif i <= 9007199254740991:  # max safe integer in JavaScript
        return i
    elif i < 18446744073709551615:
        return f"{i}"
    else:
        return "18446744073709551615"


@dataclass(frozen=True)
class _Uint64Adapter(_AbstractIntAdapter):
    def to_json_expr(self, in_expr: ExprLike, readable: bool) -> Expr:
        return Expr.join(
            Expr.local("uint64_to_json", _uint64_to_json), "(", in_expr, ")"
        )

    def get_type(self) -> reflection.Type:
        return reflection.PrimitiveType(
            kind="primitive",
            value="uint64",
        )


INT32_ADAPTER: Final[TypeAdapter] = _Int32Adapter()
INT64_ADAPTER: Final[TypeAdapter] = _Int64Adapter()
UINT64_ADAPTER: Final[TypeAdapter] = _Uint64Adapter()


@dataclass(frozen=True)
class _AbstractFloatAdapter(AbstractPrimitiveAdapter):
    """Type adapter implementation for float32 and float64."""

    def default_expr(self) -> ExprLike:
        return "0.0"

    def to_frozen_expr(self, arg_expr: ExprLike) -> Expr:
        return Expr.join("(", arg_expr, " + 0.0)")

    def is_not_default_expr(self, arg_expr: ExprLike, attr_expr: ExprLike) -> ExprLike:
        return arg_expr

    def to_json_expr(self, in_expr: ExprLike, readable: bool) -> ExprLike:
        return in_expr

    def from_json_expr(self, json_expr: ExprLike) -> Expr:
        return Expr.join("(", json_expr, " + 0.0)")


@dataclass(frozen=True)
class _Float32Adapter(_AbstractFloatAdapter):
    """Type adapter implementation for float32."""

    def get_type(self) -> reflection.Type:
        return reflection.PrimitiveType(
            kind="primitive",
            value="float32",
        )


@dataclass(frozen=True)
class _Float64Adapter(_AbstractFloatAdapter):
    """Type adapter implementation for float32."""

    def get_type(self) -> reflection.Type:
        return reflection.PrimitiveType(
            kind="primitive",
            value="float64",
        )


FLOAT32_ADAPTER: Final[TypeAdapter] = _Float32Adapter()
FLOAT64_ADAPTER: Final[TypeAdapter] = _Float64Adapter()


class _TimestampAdapter(AbstractPrimitiveAdapter):
    def default_expr(self) -> Expr:
        return Expr.local("_EPOCH", Timestamp.EPOCH)

    def to_frozen_expr(self, arg_expr: ExprLike) -> Expr:
        timestamp_local = Expr.local("Timestamp", Timestamp)
        return Expr.join(
            "(",
            arg_expr,
            " if ",
            arg_expr,
            ".__class__ is ",
            timestamp_local,
            " else ",
            timestamp_local,
            "(unix_millis=",
            arg_expr,
            ".unix_millis))",
        )

    def is_not_default_expr(self, arg_expr: ExprLike, attr_expr: ExprLike) -> Expr:
        return Expr.join(arg_expr, ".unix_millis")

    def to_json_expr(self, in_expr: ExprLike, readable: bool) -> Expr:
        if readable:
            return Expr.join(in_expr, "._trj()")
        else:
            return Expr.join(in_expr, ".unix_millis")

    def from_json_expr(self, json_expr: ExprLike) -> Expr:
        fn = Expr.local("_timestamp_from_json", _timestamp_from_json)
        return Expr.join(fn, "(", json_expr, ")")

    def get_type(self) -> reflection.Type:
        return reflection.PrimitiveType(
            kind="primitive",
            value="timestamp",
        )


def _timestamp_from_json(json: Any) -> Timestamp:
    if json.__class__ is int or isinstance(json, int):
        return Timestamp(unix_millis=json)
    else:
        return Timestamp(unix_millis=json["unix_millis"])


TIMESTAMP_ADAPTER: Final[TypeAdapter] = _TimestampAdapter()


class _StringAdapter(AbstractPrimitiveAdapter):
    def default_expr(self) -> ExprLike:
        return '""'

    def to_frozen_expr(self, arg_expr: ExprLike) -> Expr:
        return Expr.join("('' + ", arg_expr, ")")

    def is_not_default_expr(self, arg_expr: ExprLike, attr_expr: ExprLike) -> ExprLike:
        return arg_expr

    def to_json_expr(self, in_expr: ExprLike, readable: bool) -> ExprLike:
        return in_expr

    def from_json_expr(self, json_expr: ExprLike) -> Expr:
        return Expr.join("('' + (", json_expr, " or ''))")

    def get_type(self) -> reflection.Type:
        return reflection.PrimitiveType(
            kind="primitive",
            value="string",
        )


STRING_ADAPTER: Final[TypeAdapter] = _StringAdapter()


class _BytesAdapter(AbstractPrimitiveAdapter):
    _fromhex_fn: Final = bytes.fromhex

    def default_expr(self) -> ExprLike:
        return 'b""'

    def to_frozen_expr(self, arg_expr: ExprLike) -> Expr:
        return Expr.join("(b'' + ", arg_expr, ")")

    def is_not_default_expr(self, arg_expr: ExprLike, attr_expr: ExprLike) -> ExprLike:
        return arg_expr

    def to_json_expr(
        self,
        in_expr: ExprLike,
        readable: bool,
    ) -> Expr:
        return Expr.join(in_expr, ".hex()")

    def from_json_expr(self, json_expr: ExprLike) -> Expr:
        return Expr.join(
            Expr.local("fromhex", _BytesAdapter._fromhex_fn), "(", json_expr, ' or "")'
        )

    def get_type(self) -> reflection.Type:
        return reflection.PrimitiveType(
            kind="primitive",
            value="bytes",
        )


BYTES_ADAPTER: Final[TypeAdapter] = _BytesAdapter()



from typing import Any, Union

from soia._impl import arrays, enums, method, optionals, primitives, structs
from soia._impl.serializer import make_serializer
from soia._impl.type_adapter import TypeAdapter

from soia import _spec

RecordAdapter = Union[structs.StructAdapter, enums.EnumAdapter]


_record_id_to_adapter: dict[str, RecordAdapter] = {}


def init_module(
    records: tuple[_spec.Record, ...],
    methods: tuple[_spec.Method, ...],
    constants: tuple[_spec.Constant, ...],
    globals: dict[str, Any],
    # For testing
    record_id_to_adapter: dict[str, RecordAdapter] = _record_id_to_adapter,
) -> None:
    def resolve_type(type: _spec.Type) -> TypeAdapter:
        if isinstance(type, _spec.PrimitiveType):
            if type == _spec.PrimitiveType.BOOL:
                return primitives.BOOL_ADAPTER
            elif type == _spec.PrimitiveType.BYTES:
                return primitives.BYTES_ADAPTER
            elif type == _spec.PrimitiveType.FLOAT32:
                return primitives.FLOAT32_ADAPTER
            elif type == _spec.PrimitiveType.FLOAT64:
                return primitives.FLOAT64_ADAPTER
            elif type == _spec.PrimitiveType.INT32:
                return primitives.INT32_ADAPTER
            elif type == _spec.PrimitiveType.INT64:
                return primitives.INT64_ADAPTER
            elif type == _spec.PrimitiveType.STRING:
                return primitives.STRING_ADAPTER
            elif type == _spec.PrimitiveType.TIMESTAMP:
                return primitives.TIMESTAMP_ADAPTER
            elif type == _spec.PrimitiveType.UINT64:
                return primitives.UINT64_ADAPTER
        elif isinstance(type, _spec.ArrayType):
            return arrays.get_array_adapter(
                resolve_type(type.item),
                type.key_attributes,
            )
        elif isinstance(type, _spec.OptionalType):
            return optionals.get_optional_adapter(resolve_type(type.other))
        elif isinstance(type, str):
            # A record id.
            return record_id_to_adapter[type]

    module_adapters: list[RecordAdapter] = []
    for record in records:
        if record.id in record_id_to_adapter:
            raise AssertionError(record.id)
        adapter: RecordAdapter
        if isinstance(record, _spec.Struct):
            adapter = structs.StructAdapter(record)
        else:
            adapter = enums.EnumAdapter(record)
        module_adapters.append(adapter)
        record_id_to_adapter[record.id] = adapter
    # Once all the adapters of the module have been created, we can finalize them.
    for adapter in module_adapters:
        adapter.finalize(resolve_type)
        gen_class = adapter.gen_class
        # Add the class name to either globals() if the record is defined at the top
        # level, or the parent class otherwise.
        record_id = _spec.RecordId.parse(adapter.spec.id)
        parent_id = record_id.parent
        class_name = adapter.spec.class_name
        if parent_id:
            parent_adapter = record_id_to_adapter[parent_id.record_id]
            setattr(parent_adapter.gen_class, class_name, gen_class)
            gen_class._parent_class = parent_adapter.gen_class
        else:
            globals[class_name] = gen_class
            gen_class._parent_class = None
        gen_class.SERIALIZER = make_serializer(adapter)

    # Now that al the classes have been initialized, create the methods.
    for method_spec in methods:
        var_name = method_spec._var_name or method_spec.name
        request_serializer = make_serializer(resolve_type(method_spec.request_type))
        response_serializer = make_serializer(resolve_type(method_spec.response_type))
        globals[var_name] = method.Method(
            name=method_spec.name,
            number=method_spec.number,
            request_serializer=request_serializer,
            response_serializer=response_serializer,
        )
        del var_name, request_serializer, response_serializer

    # Create the constants.
    for constant in constants:
        serializer = make_serializer(resolve_type(constant.type))
        globals[constant.name] = serializer.from_json_code(constant.json_code)
        del serializer



import itertools
from dataclasses import dataclass
from typing import Any, Callable, Iterable, Optional, Sequence, Union


def make_function(
    name: str,
    params: Sequence[Union[str, "Param"]],
    body: Sequence[Union[str, "Line"]],
) -> Callable:
    params = [Param(p, default=None) if isinstance(p, str) else p for p in params]

    def make_locals() -> _Locals:
        # First, collect all the (key, value) pairs.
        # There may be both key duplicates and value duplicates. They will be resolved
        # later.
        all_pieces: list[Union[str, _Local]] = []
        for param in params:
            if param._default is not None:
                all_pieces.extend(param._default._pieces)
        for line in body:
            if isinstance(line, LineSpan):
                all_pieces.extend(line._pieces)
        all_locals = (p for p in all_pieces if isinstance(p, _Local))

        locals: dict[str, Any] = {}
        reversed: dict[int, str] = {}
        prefix_to_count: dict[str, int] = {}
        for local in all_locals:
            name = local.name
            value = local.value
            if id(value) in reversed:
                # We already have this value, either under the same name or under a
                # different name. Either way, we can reuse the name we already have.
                continue
            if name.endswith("?"):
                # Replace the "?" suffix with a number obtained from a sequence so there
                # is no name conflict.
                prefix = name[0:-1]
                count = prefix_to_count.get(prefix, 0)
                prefix_to_count[prefix] = count + 1
                name = prefix + str(count)
            if name in locals:
                raise ValueError(f"Duplicate local: name={name}; value={value}")
            locals[name] = value
            reversed[id(value)] = name

        return _Locals(locals, reversed)

    locals = make_locals()

    def line_to_code(ln: Union[str, Line]) -> str:
        if isinstance(ln, str):
            return ln
        return ln._to_code(locals)

    body_str = "\n    ".join(line_to_code(ln) for ln in body) if body else "pass"
    text = f"""
def __create_function__({', '.join(locals.locals.keys())}):
  def {name}({', '.join(p._to_code(locals) for p in params)}):
    {body_str}
  return {name}
"""
    ns = {}
    exec(text, None, ns)
    return ns["__create_function__"](**locals.locals)


@dataclass(frozen=True)
class LineSpan:
    "An immutable span within a line of Python code."

    _pieces: tuple[Union[str, "_Local"], ...]

    @staticmethod
    def join(*spans: Union[str, "LineSpan"], separator: str = "") -> "LineSpan":
        def get_pieces(
            span: Union[str, "LineSpan"],
        ) -> tuple[Union[str, "_Local"], ...]:
            if isinstance(span, LineSpan):
                return span._pieces
            return ((span),)

        # Remove empty strings.
        spans = tuple(s for s in spans if s)
        if len(spans) == 0:
            return _EMPTY_LINE_SPAN
        elif len(spans) == 1:
            only_span = spans[0]
            if isinstance(only_span, LineSpan):
                return only_span
            return LineSpan((only_span,))
        if separator:
            pieces = itertools.chain(
                get_pieces(spans[0]),
                *itertools.chain(
                    itertools.chain((separator,), get_pieces(s)) for s in spans[1:]
                ),
            )
        else:
            pieces = itertools.chain(*(get_pieces(s) for s in spans))
        return LineSpan(tuple(pieces))

    @staticmethod
    def local(name: str, value: Any):
        return LineSpan((_Local(name, value),))

    def _to_code(self, locals: "_Locals") -> str:
        ret = ""
        for piece in self._pieces:
            if isinstance(piece, _Local):
                # Look up the resolved name for the value.
                # It migth be different from `piece.name` in some edge cases.
                ret += locals.reversed[id(piece.value)]
            else:
                ret += piece
        return ret


_EMPTY_LINE_SPAN = LineSpan(())


LineSpanLike = Union[str, LineSpan]


# A type alias to use when the line span is a Python expression.
Expr = LineSpan
ExprLike = LineSpanLike


# A type alias to use when the line span is the whole line.
Line = LineSpan


# An immutable sequence of lines.
BodySpan = tuple[Union[str, Line], ...]


# A type alias to use when the lines constitute the whole body of a function.
Body = BodySpan


class BodyBuilder:
    """Arnold Schwarzenegger."""

    _lines: list[LineSpan]

    def __init__(self):
        self._lines = []

    def append_ln(self, *spans: Union[str, LineSpan]):
        self._lines.append(Line.join(*spans))

    def extend(self, other: Iterable[Union[str, Line]], indent: str = ""):
        self._lines.extend(Line.join(indent, ln) for ln in other)
        return self

    def build(self) -> Body:
        return tuple(self._lines)


@dataclass(frozen=True)
class Param:
    name: str
    default: Optional[ExprLike] = None

    def _to_code(self, locals: "_Locals") -> str:
        if self._default is None:
            return self.name
        return f"{self.name}={self._default._to_code(locals)}"

    @property
    def _default(self) -> Optional[Expr]:
        if self.default is None:
            return None
        else:
            return Expr.join(self.default)


Params = list[Union[str, Param]]


@dataclass(frozen=True)
class _Local:
    name: str
    value: Any


@dataclass(frozen=True)
class _Locals:
    locals: dict[str, Any]
    reversed: dict[Any, str]



import http.client
import re
from typing import Any, Final, Mapping, Protocol
from urllib.parse import urlparse

from soia._impl.method import Method, Request, Response


class ServiceClient:
    _scheme: Final[str]
    _host: Final[str]  # May include the port
    _path: Final[str]

    def __init__(self, url: str):
        parsed_url = urlparse(url)
        if parsed_url.query:
            raise ValueError("Service URL must not contain a query string")
        scheme = parsed_url.scheme
        if scheme not in ["http", "https"]:
            raise ValueError("Service URL must start with http:// or https://")
        self._scheme = scheme
        self._host = parsed_url.netloc
        self._path = parsed_url.path

    def invoke_remote(
        self,
        method: Method[Request, Response],
        request: Request,
        headers: Mapping[str, str] = {},
        *,
        res_headers: list[tuple[str, str]] | None = None,
        timeout_secs: float | None = None,
    ) -> Response:
        """Invokes the given method on the remote server through an RPC."""

        request_json = method.request_serializer.to_json_code(request)
        body = ":".join(
            [
                method.name,
                str(method.number),
                "",
                request_json,
            ]
        )
        headers = {
            **headers,
            "Content-Type": "text/plain; charset=utf-8",
            "Content-Length": str(len(body)),
        }
        connection_options: dict[str, Any] = {}
        if timeout_secs is not None:
            connection_options["timeout"] = timeout_secs
        if self._scheme == "https":
            conn = http.client.HTTPSConnection(self._host, **connection_options)
        else:
            conn = http.client.HTTPConnection(self._host, **connection_options)
        try:
            conn.request(
                "POST",
                self._path,
                body=body,
                headers=headers,
            )
            response = conn.getresponse()
            if res_headers is not None:
                res_headers.clear()
                res_headers.extend(response.getheaders())
            status_code = response.status
            content_type = response.getheader("Content-Type") or ""
            response_data = response.read().decode("utf-8", errors="ignore")
        finally:
            conn.close()
        if status_code in range(200, 300):
            return method.response_serializer.from_json_code(response_data)
        else:
            message = f"HTTP status {status_code}"
            if re.match(r"text/plain\b", content_type):
                message = f"{message}: {response_data}"
            raise RuntimeError(message)

    async def invoke_remote_async(
        self,
        aiohttp_client_session: "_AiohttpClientSession",
        method: Method[Request, Response],
        request: Request,
        headers: Mapping[str, str] = {},
        *,
        res_headers: list[tuple[str, str]] | None = None,
    ) -> Response:
        """
        Asynchronously invokes the given method on the remote server through an RPC.

        Usage:
            import aiohttp

            async with aiohttp.ClientSession() as session:
                response = await service_client.invoke_remote_async(
                    session,
                    method,
                    request,
                    headers,
                )

        With timeout:
            timeout = aiohttp.ClientTimeout(total=30)
            async with aiohttp.ClientSession(timeout=timeout) as session:
                response = await service_client.invoke_remote_async(
                    session,
                    method,
                    request,
                    headers,
                )
        """

        request_json = method.request_serializer.to_json_code(request)
        body = ":".join(
            [
                method.name,
                str(method.number),
                "",
                request_json,
            ]
        )
        
        request_headers = {
            **headers,
            "Content-Type": "text/plain; charset=utf-8",
            "Content-Length": str(len(body)),
        }

        # Build the URL
        url = f"{self._scheme}://{self._host}{self._path}"

        async with aiohttp_client_session.post(
            url,
            data=body,
            headers=request_headers,
        ) as response:
            if res_headers is not None:
                res_headers.clear()
                res_headers.extend(response.headers.items())

            status_code = response.status
            content_type = response.headers.get("Content-Type", "")
            response_data = await response.text(encoding="utf-8", errors="ignore")

        if status_code in range(200, 300):
            return method.response_serializer.from_json_code(response_data)
        else:
            message = f"HTTP status {status_code}"
            if re.match(r"text/plain\b", content_type):
                message = f"{message}: {response_data}"
            raise RuntimeError(message)


class _AiohttpClientSession(Protocol):
    def post(
        self,
        url: str,
        *,
        data: str,
        headers: Mapping[str, str],
    ) -> Any:
        ...



from collections.abc import Callable
from dataclasses import FrozenInstanceError
from typing import Generic, Optional
from weakref import WeakValueDictionary

from soia._impl.function_maker import Any, Expr, ExprLike, Line, make_function
from soia._impl.keyed_items import Item, Key, KeyedItems
from soia._impl.type_adapter import TypeAdapter

from soia import _spec, reflection


def get_array_adapter(
    item_adapter: TypeAdapter,
    key_attributes: tuple[str, ...],
) -> TypeAdapter:
    if key_attributes:
        default_expr = item_adapter.default_expr()
        listuple_class = _new_keyed_items_class(key_attributes, default_expr)
    else:
        listuple_class = _new_listuple_class()
    array_adapter = _ArrayAdapter(item_adapter, listuple_class, key_attributes)
    return _item_to_array_adapter.setdefault(
        (item_adapter, key_attributes), array_adapter
    )


class _ArrayAdapter(TypeAdapter):
    __slots__ = (
        "item_adapter",
        "listuple_class",
        "key_attributes",
        "empty_listuple",
    )

    item_adapter: TypeAdapter
    listuple_class: type
    empty_listuple: tuple[()]

    def __init__(
        self,
        item_adapter: TypeAdapter,
        listuple_class: type,
        key_attributes: tuple[str, ...],
    ):
        self.item_adapter = item_adapter
        self.listuple_class = listuple_class
        self.key_attributes = key_attributes
        self.empty_listuple = listuple_class()

    def default_expr(self) -> ExprLike:
        return "()"

    def to_frozen_expr(self, arg_expr: ExprLike) -> Expr:
        listuple_class_local = Expr.local("_lstpl?", self.listuple_class)
        empty_listuple_local = Expr.local("_emp?", self.empty_listuple)
        return Expr.join(
            "(",
            arg_expr,
            " if ",
            arg_expr,
            ".__class__ is ",
            listuple_class_local,
            " else (",
            listuple_class_local,
            "([",
            self.item_adapter.to_frozen_expr("_e"),
            " for _e in ",
            arg_expr,
            "]) or ",
            empty_listuple_local,
            "))",
        )

    def is_not_default_expr(self, arg_expr: ExprLike, attr_expr: ExprLike) -> ExprLike:
        # Can't use arg_expr, an empty iterable is not guaranteed to evaluate to False.
        return attr_expr

    def to_json_expr(self, in_expr: ExprLike, readable: bool) -> ExprLike:
        item_to_json = self.item_adapter.to_json_expr("_e", readable)
        if Expr.join(item_to_json) == Expr.join("_e"):
            return in_expr
        return Expr.join(
            "[",
            item_to_json,
            " for _e in ",
            in_expr,
            "]",
        )

    def from_json_expr(self, json_expr: ExprLike) -> Expr:
        listuple_class_local = Expr.local("_lstpl?", self.listuple_class)
        empty_listuple_local = Expr.local("_emp?", self.empty_listuple)
        return Expr.join(
            listuple_class_local,
            "([",
            self.item_adapter.from_json_expr("_e"),
            " for _e in ",
            json_expr,
            "] or ",
            empty_listuple_local,
            ")",
        )

    def finalize(
        self,
        resolve_type_fn: Callable[[_spec.Type], "TypeAdapter"],
    ) -> None:
        self.item_adapter.finalize(resolve_type_fn)

    def get_type(self) -> reflection.Type:
        return reflection.ArrayType(
            kind="array",
            value=reflection.ArrayType.Array(
                item=self.item_adapter.get_type(),
                key_chain=self.key_attributes,
            ),
        )

    def register_records(
        self,
        registry: dict[str, reflection.Record],
    ) -> None:
        self.item_adapter.register_records(registry)


_ItemAndKeyAttributes = tuple[TypeAdapter, tuple[str, ...]]
_ItemToArrayAdapter = WeakValueDictionary[_ItemAndKeyAttributes, _ArrayAdapter]

_item_to_array_adapter: _ItemToArrayAdapter = WeakValueDictionary()


def _new_listuple_class() -> type:
    class Listuple(Generic[Item], tuple[Item, ...]):
        __slots__ = ()

    return Listuple


def _new_keyed_items_class(key_attributes: tuple[str, ...], default_expr: ExprLike):
    key_items = make_function(
        name="key_items",
        params=["items"],
        body=[
            "ret = {}",
            "for item in items:",
            f"  ret[item.{'.'.join(key_attributes)}] = item",
            "return ret",
        ],
    )

    default = make_function(
        name="get_default",
        params="",
        body=[Line.join("return ", default_expr)],
    )()

    class KeyedItemsImpl(KeyedItems[Item, Key]):
        # nonempty __slots__ not supported for subtype of 'tuple'

        _key_to_item: dict[Key, Item]

        def find(self, key: Key) -> Optional[Item]:
            try:
                key_to_item = self._key_to_item
            except AttributeError:
                key_to_item = key_items(self)
                object.__setattr__(self, "_key_to_item", key_to_item)
            return key_to_item.get(key)

        def find_or_default(self, key: Key) -> Any:
            return self.find(key) or default

        def __setattr__(self, name: str, value: Any):
            raise FrozenInstanceError(self.__class__.__qualname__)

        def __delattr__(self, name: str):
            raise FrozenInstanceError(self.__class__.__qualname__)

    return KeyedItemsImpl



import unittest

from soia import Timestamp, array_serializer, optional_serializer, primitive_serializer


class TimestampTestCase(unittest.TestCase):
    def test_primitive_serializers(self):
        self.assertEqual(primitive_serializer("bool").to_json_code(True), "1")
        self.assertEqual(
            primitive_serializer("bool").to_json_code(True, readable=True),
            "true",
        )
        self.assertEqual(primitive_serializer("bool").from_json_code("true"), True)
        self.assertEqual(primitive_serializer("bool").from_json_code("false"), False)
        self.assertEqual(primitive_serializer("bool").from_json_code("0"), False)

        self.assertEqual(primitive_serializer("int32").to_json_code(7), "7")
        self.assertEqual(
            primitive_serializer("int32").to_json_code(7, readable=True), "7"
        )
        self.assertEqual(
            primitive_serializer("int32").to_json_code(2147483648), "2147483647"
        )
        self.assertEqual(
            primitive_serializer("int32").to_json_code(-2147483649), "-2147483648"
        )
        self.assertEqual(primitive_serializer("int32").from_json_code("0"), 0)
        self.assertEqual(primitive_serializer("int32").from_json_code('"7"'), 7)
        self.assertEqual(primitive_serializer("int32").from_json_code("3.14"), 3)
        self.assertEqual(primitive_serializer("int32").from_json_code("-3.14"), -3)

        self.assertEqual(
            primitive_serializer("int64").to_json_code(2147483648), "2147483648"
        )
        self.assertEqual(
            primitive_serializer("int64").to_json_code(9007199254740991),
            "9007199254740991",
        )
        self.assertEqual(
            primitive_serializer("int64").to_json_code(9007199254740992),
            '"9007199254740992"',
        )
        self.assertEqual(
            primitive_serializer("int64").to_json_code(9223372036854775808),
            '"9223372036854775807"',
        )
        self.assertEqual(
            primitive_serializer("int64").to_json_code(-9007199254740991),
            "-9007199254740991",
        )
        self.assertEqual(
            primitive_serializer("int64").to_json_code(-9007199254740992),
            '"-9007199254740992"',
        )
        self.assertEqual(
            primitive_serializer("int64").to_json_code(-9223372036854775809),
            '"-9223372036854775808"',
        )
        self.assertEqual(primitive_serializer("int64").from_json_code("0"), 0)
        self.assertEqual(primitive_serializer("int64").from_json_code('"7"'), 7)
        self.assertEqual(primitive_serializer("int64").from_json_code("3.14"), 3)
        self.assertEqual(primitive_serializer("int64").from_json_code("-3.14"), -3)

        self.assertEqual(
            primitive_serializer("uint64").to_json_code(2147483648),
            "2147483648",
        )
        self.assertEqual(
            primitive_serializer("uint64").to_json_code(-1),
            "0",
        )
        self.assertEqual(
            primitive_serializer("uint64").to_json_code(9007199254740991),
            "9007199254740991",
        )
        self.assertEqual(
            primitive_serializer("uint64").to_json_code(9007199254740992),
            '"9007199254740992"',
        )
        self.assertEqual(primitive_serializer("uint64").from_json_code("0"), 0)
        self.assertEqual(primitive_serializer("uint64").from_json_code('"7"'), 7)
        self.assertEqual(primitive_serializer("uint64").from_json_code("3.14"), 3)
        self.assertEqual(primitive_serializer("uint64").from_json_code("-3.14"), -3)

        self.assertEqual(primitive_serializer("float32").to_json_code(3.14), "3.14")
        self.assertEqual(
            primitive_serializer("float32").to_json_code(float("inf")), "Infinity"
        )
        self.assertEqual(
            primitive_serializer("float32").to_json_code(-float("inf")), "-Infinity"
        )
        self.assertEqual(
            primitive_serializer("float32").to_json_code(float("nan")), "NaN"
        )
        self.assertEqual(primitive_serializer("float32").from_json_code("3.14"), 3.14)
        self.assertEqual(
            primitive_serializer("float32").from_json_code("Infinity"), float("inf")
        )
        self.assertEqual(
            primitive_serializer("float32").from_json_code("-Infinity"), -float("inf")
        )
        self.assertNotEqual(
            primitive_serializer("float32").from_json_code("NaN"),
            primitive_serializer("float32").from_json_code("NaN"),
        )

        self.assertEqual(primitive_serializer("float64").to_json_code(3.14), "3.14")
        self.assertEqual(
            primitive_serializer("float64").to_json_code(3.14, readable=True),
            "3.14",
        )

        self.assertEqual(
            primitive_serializer("timestamp").to_json_code(
                Timestamp.from_unix_millis(3)
            ),
            "3",
        )
        self.assertEqual(
            primitive_serializer("timestamp").to_json_code(
                Timestamp.from_unix_millis(3), readable=True
            ),
            '{\n  "unix_millis": 3,\n  "formatted": "1970-01-01T00:00:00.003000Z"\n}',
        )
        self.assertEqual(
            primitive_serializer("timestamp").from_json_code("3"),
            Timestamp.from_unix_millis(3),
        )
        self.assertEqual(
            primitive_serializer("timestamp").from_json_code(
                '{\n  "unix_millis": 3,\n  "formatted": "FOO",\n  "bar": true\n}'
            ),
            Timestamp.from_unix_millis(3),
        )

    def test_array_serializer(self):
        self.assertEqual(
            array_serializer(primitive_serializer("bool")).to_json_code((True, False)),
            "[1,0]",
        )
        self.assertEqual(
            array_serializer(primitive_serializer("bool")).to_json_code(
                (True, False), readable=True
            ),
            "[\n  true,\n  false\n]",
        )

    def test_optional_serializer(self):
        self.assertEqual(
            optional_serializer(primitive_serializer("bool")).to_json_code(True),
            "1",
        )
        self.assertEqual(
            optional_serializer(primitive_serializer("bool")).to_json_code(
                True, readable=True
            ),
            "true",
        )
        self.assertEqual(
            optional_serializer(primitive_serializer("bool")).to_json_code(None),
            "null",
        )



import enum
from dataclasses import dataclass
from typing import Optional, Union


class PrimitiveType(enum.Enum):
    BOOL = enum.auto()
    INT32 = enum.auto()
    INT64 = enum.auto()
    UINT64 = enum.auto()
    FLOAT32 = enum.auto()
    FLOAT64 = enum.auto()
    TIMESTAMP = enum.auto()
    STRING = enum.auto()
    BYTES = enum.auto()


@dataclass(frozen=True)
class ArrayType:
    item: "Type"
    key_attributes: tuple[str, ...] = ()


@dataclass(frozen=True)
class OptionalType:
    other: "Type"


Type = Union[PrimitiveType, ArrayType, OptionalType, str]


@dataclass(frozen=True)
class Field:
    """Field of a struct."""

    name: str
    number: int
    type: Type
    has_mutable_getter: bool = False
    _attribute: str = ""  # If different from 'name'

    @property
    def attribute(self):
        return self._attribute or self.name


@dataclass(frozen=True)
class Struct:
    id: str
    fields: tuple[Field, ...] = ()
    removed_numbers: tuple[int, ...] = ()
    _class_name: str = ""  # If different from the record name
    _class_qualname: str = ""  # If different from the qualified name of the record

    @property
    def class_name(self):
        return self._class_name or RecordId.parse(self.id).name

    @property
    def class_qualname(self):
        return self._class_qualname or RecordId.parse(self.id).qualname


@dataclass(frozen=True)
class ConstantField:
    """Constant field of an enum."""

    name: str
    number: int
    _attribute: str = ""  # If different from 'name'

    @property
    def attribute(self):
        return self._attribute or self.name


@dataclass(frozen=True)
class ValueField:
    """Value field of an enum."""

    name: str
    number: int
    type: Type


@dataclass(frozen=True)
class Enum:
    id: str
    constant_fields: tuple[ConstantField, ...] = ()
    value_fields: tuple[ValueField, ...] = ()
    removed_numbers: tuple[int, ...] = ()
    _class_name: str = ""  # If different from the record name
    _class_qualname: str = ""  # If different from the qualified name of the record

    @property
    def class_name(self):
        return self._class_name or RecordId.parse(self.id).name

    @property
    def class_qualname(self):
        return self._class_qualname or RecordId.parse(self.id).qualname


Record = Union[Struct, Enum]


@dataclass(frozen=True)
class RecordId:
    record_id: str
    module_path: str
    name: str
    qualname: str
    name_parts: tuple[str, ...]

    @staticmethod
    def parse(record_id: str) -> "RecordId":
        colon_index = record_id.rfind(":")
        module_path = record_id[0:colon_index]
        qualname = record_id[colon_index + 1 :]
        name_parts = tuple(qualname.split("."))
        return RecordId(
            record_id=record_id,
            module_path=module_path,
            name=name_parts[-1],
            qualname=qualname,
            name_parts=name_parts,
        )

    @property
    def parent(self) -> Optional["RecordId"]:
        if len(self.name_parts) == 1:
            return None
        parent_name_parts = self.name_parts[0:-1]
        parent_qualname = ".".join(parent_name_parts)
        return RecordId(
            record_id=f"{self.module_path}:{parent_qualname}",
            module_path=self.module_path,
            name=parent_name_parts[-1],
            qualname=parent_qualname,
            name_parts=parent_name_parts,
        )


@dataclass(frozen=True)
class Method:
    name: str
    number: int
    request_type: Type
    response_type: Type
    _var_name: str = ""  # If different from 'name'


@dataclass(frozen=True)
class Constant:
    name: str
    type: Type
    json_code: str



from typing import Final, Literal, TypeVar, overload

from soia._impl import primitives
from soia._impl.arrays import get_array_adapter
from soia._impl.optionals import get_optional_adapter
from soia._impl.serializer import Serializer, make_serializer
from soia._impl.timestamp import Timestamp

Item = TypeVar("Item")
Other = TypeVar("Other")


def array_serializer(item_serializer: Serializer[Item]) -> Serializer[tuple[Item, ...]]:
    return make_serializer(get_array_adapter(item_serializer._adapter, ()))


def optional_serializer(
    other_serializer: Serializer[Other],
) -> Serializer[Other | None]:
    return make_serializer(get_optional_adapter(other_serializer._adapter))


@overload
def primitive_serializer(primitive: Literal["bool"]) -> Serializer[bool]: ...
@overload
def primitive_serializer(primitive: Literal["int32"]) -> Serializer[int]: ...
@overload
def primitive_serializer(primitive: Literal["int64"]) -> Serializer[int]: ...
@overload
def primitive_serializer(primitive: Literal["uint64"]) -> Serializer[int]: ...
@overload
def primitive_serializer(primitive: Literal["float32"]) -> Serializer[float]: ...
@overload
def primitive_serializer(primitive: Literal["float64"]) -> Serializer[float]: ...
@overload
def primitive_serializer(primitive: Literal["timestamp"]) -> Serializer[Timestamp]: ...
@overload
def primitive_serializer(primitive: Literal["string"]) -> Serializer[str]: ...
@overload
def primitive_serializer(primitive: Literal["bytes"]) -> Serializer[bytes]: ...


def primitive_serializer(
    primitive: (
        Literal["bool"]
        | Literal["int32"]
        | Literal["int64"]
        | Literal["uint64"]
        | Literal["float32"]
        | Literal["float64"]
        | Literal["timestamp"]
        | Literal["string"]
        | Literal["bytes"]
    ),
) -> (
    Serializer[bool]
    | Serializer[int]
    | Serializer[float]
    | Serializer[Timestamp]
    | Serializer[str]
    | Serializer[bytes]
):
    return _PRIMITIVE_TO_SERIALIZER[primitive]


_PRIMITIVE_TO_SERIALIZER: Final = {
    "bool": make_serializer(primitives.BOOL_ADAPTER),
    "int32": make_serializer(primitives.INT32_ADAPTER),
    "int64": make_serializer(primitives.INT64_ADAPTER),
    "uint64": make_serializer(primitives.UINT64_ADAPTER),
    "float32": make_serializer(primitives.FLOAT32_ADAPTER),
    "float64": make_serializer(primitives.FLOAT64_ADAPTER),
    "timestamp": make_serializer(primitives.TIMESTAMP_ADAPTER),
    "string": make_serializer(primitives.STRING_ADAPTER),
    "bytes": make_serializer(primitives.BYTES_ADAPTER),
}



import inspect
import json
from collections.abc import Awaitable, Callable
from dataclasses import dataclass
from typing import Any, Generic, Literal, TypeVar, Union, cast

from soia._impl.method import Method, Request, Response

RequestHeaders = TypeVar("RequestHeaders")

ResponseHeaders = TypeVar("ResponseHeaders")


@dataclass(frozen=True)
class _MethodImpl(Generic[Request, Response, RequestHeaders, ResponseHeaders]):
    method: Method[Request, Response]
    impl: Callable[
        # Parameters
        [Request, RequestHeaders, ResponseHeaders],
        # Return type
        Union[Response, Awaitable[Response]],
    ]


@dataclass(frozen=True)
class RawServiceResponse:
    data: str
    type: Literal["ok-json", "bad-request", "server-error"]

    @property
    def status_code(self):
        if self.type == "ok-json":
            return 200
        elif self.type == "bad-request":
            return 400
        elif self.type == "server-error":
            return 500
        else:
            raise TypeError(f"Unknown response type: {self.type}")

    @property
    def content_type(self):
        if self.type == "ok-json":
            return "application/json"
        elif self.type == "bad-request" or self.type == "server-error":
            return "text/plain; charset=utf-8"
        else:
            raise TypeError(f"Unknown response type: {self.type}")


@dataclass()
class _HandleRequestFlow(Generic[Request, Response, RequestHeaders, ResponseHeaders]):
    req_body: str
    req_headers: RequestHeaders
    res_headers: ResponseHeaders
    number_to_method_impl: dict[
        int, _MethodImpl[Any, Any, RequestHeaders, ResponseHeaders]
    ]
    _format: str = ""

    def run(self) -> RawServiceResponse:
        req_impl_pair_or_raw_response = self._parse_request()
        if isinstance(req_impl_pair_or_raw_response, RawServiceResponse):
            return req_impl_pair_or_raw_response
        req, method_impl = req_impl_pair_or_raw_response
        try:
            res = method_impl.impl(req, self.req_headers, self.res_headers)
        except Exception as e:
            return RawServiceResponse(f"server error: {e}", "server-error")
        if inspect.isawaitable(res):
            raise TypeError("Method implementation must be synchronous")
        return self._response_to_json(res, method_impl)

    async def run_async(self) -> RawServiceResponse:
        req_impl_pair_or_raw_response = self._parse_request()
        if isinstance(req_impl_pair_or_raw_response, RawServiceResponse):
            return req_impl_pair_or_raw_response
        req, method_impl = req_impl_pair_or_raw_response
        try:
            res: Any = method_impl.impl(req, self.req_headers, self.res_headers)
            if inspect.isawaitable(res):
                res = await res
        except Exception as e:
            return RawServiceResponse(f"server error: {e}", "server-error")
        return self._response_to_json(res, method_impl)

    def _parse_request(
        self,
    ) -> Union[
        tuple[Any, _MethodImpl[Request, Response, RequestHeaders, ResponseHeaders]],
        RawServiceResponse,
    ]:
        if self.req_body == "list":

            def method_to_json(method: Method) -> Any:
                return {
                    "method": method.name,
                    "number": method.number,
                    "request": method.request_serializer.type_descriptor.as_json(),
                    "response": method.response_serializer.type_descriptor.as_json(),
                }

            json_code = json.dumps(
                {
                    "methods": [
                        method_to_json(method_impl.method)
                        for method_impl in self.number_to_method_impl.values()
                    ]
                },
                indent=2,
            )
            return RawServiceResponse(json_code, "ok-json")

        parts = self.req_body.split(":", 3)
        if len(parts) != 4:
            return RawServiceResponse(
                "bad request: invalid request format", "bad-request"
            )
        method_name = parts[0]
        method_number_str = parts[1]
        self.format = parts[2]
        request_data = parts[3]
        try:
            method_number = int(method_number_str)
        except Exception:
            return RawServiceResponse(
                "bad request: can't parse method number", "bad-request"
            )
        method_impl = self.number_to_method_impl.get(method_number)
        if not method_impl:
            return RawServiceResponse(
                f"bad request: method not found: {method_name}; number: {method_number}",
                "bad-request",
            )
        try:
            req: Any = method_impl.method.request_serializer.from_json_code(
                request_data
            )
        except Exception as e:
            return RawServiceResponse(
                f"bad request: can't parse JSON: {e}", "bad-request"
            )
        return (req, method_impl)

    def _response_to_json(
        self,
        res: Response,
        method_impl: _MethodImpl[Request, Response, RequestHeaders, ResponseHeaders],
    ) -> RawServiceResponse:
        try:
            res_json = method_impl.method.response_serializer.to_json_code(
                res, readable=(self.format == "readable")
            )
        except Exception as e:
            return RawServiceResponse(
                f"server error: can't serialize response to JSON: {e}", "server-error"
            )
        return RawServiceResponse(res_json, "ok-json")


class _ServiceImpl(Generic[RequestHeaders, ResponseHeaders]):
    _number_to_method_impl: dict[
        int, _MethodImpl[Any, Any, RequestHeaders, ResponseHeaders]
    ]

    def __init__(self):
        self._number_to_method_impl = {}

    def add_method(
        self,
        method: Method[Request, Response],
        impl: Union[
            # Sync
            Callable[[Request], Response],
            Callable[[Request, RequestHeaders], Response],
            Callable[[Request, RequestHeaders, ResponseHeaders], Response],
            # Async
            Callable[[Request], Awaitable[Response]],
            Callable[[Request, RequestHeaders], Awaitable[Response]],
            Callable[[Request, RequestHeaders, ResponseHeaders], Awaitable[Response]],
        ],
    ) -> None:
        signature = inspect.Signature.from_callable(impl)
        num_positional_params = 0
        for param in signature.parameters.values():
            if param.kind in (
                inspect.Parameter.POSITIONAL_OR_KEYWORD,
                inspect.Parameter.POSITIONAL_ONLY,
            ):
                num_positional_params += 1
            if param.kind == inspect.Parameter.VAR_POSITIONAL:
                raise ValueError("Method implementation cannot accept *args")
        if num_positional_params not in range(1, 4):
            raise ValueError(
                "Method implementation must accept 1 to 3 positional parameters"
            )

        def resolved_impl(
            req: Request, req_headers: RequestHeaders, res_headers: ResponseHeaders
        ) -> Response:
            if num_positional_params == 1:
                return cast(Callable[[Request], Response], impl)(req)
            elif num_positional_params == 2:
                return cast(Callable[[Request, RequestHeaders], Response], impl)(
                    req, req_headers
                )
            else:
                return cast(
                    Callable[[Request, RequestHeaders, ResponseHeaders], Response], impl
                )(req, req_headers, res_headers)

        number = method.number
        if number in self._number_to_method_impl:
            raise ValueError(
                f"Method with the same number already registered ({number})"
            )
        self._number_to_method_impl[number] = _MethodImpl(
            method=method,
            impl=resolved_impl,
        )

    def handle_request(
        self,
        req_body: str,
        req_headers: RequestHeaders,
        res_headers: ResponseHeaders,
    ) -> RawServiceResponse:
        flow = _HandleRequestFlow(
            req_body=req_body,
            req_headers=req_headers,
            res_headers=res_headers,
            number_to_method_impl=self._number_to_method_impl,
        )
        return flow.run()

    async def handle_request_async(
        self,
        req_body: str,
        req_headers: RequestHeaders,
        res_headers: ResponseHeaders,
    ) -> RawServiceResponse:
        flow = _HandleRequestFlow(
            req_body=req_body,
            req_headers=req_headers,
            res_headers=res_headers,
            number_to_method_impl=self._number_to_method_impl,
        )
        return await flow.run_async()


class Service(Generic[RequestHeaders, ResponseHeaders]):
    """Wraps around the implementation of a soia service on the server side.

    Usage: call '.add_method()' to register method implementations, then call
    '.handle_request()' from the function called by your web framework when an
    HTTP request is received at your service's endpoint.

    Example with Flask:

        from flask import Response, request
        from werkzeug.datastructures import Headers


        s = soia.Service[Headers, Headers]()
        s.add_method(...)
        s.add_method(...)

        @app.route("/myapi", methods=["GET", "POST"])
        def myapi():
            if request.method == "POST":
                req_body = request.get_data(as_text=True)
            else:
                query_string = request.query_string.decode("utf-8")
                req_body = urllib.parse.unquote(query_string)
            req_headers = request.headers
            res_headers = Headers()
            raw_response = s.handle_request(req_body, req_headers, res_headers)
            return Response(
                raw_response.data,
                status=raw_response.status_code,
                content_type=raw_response.content_type,
                headers=res_headers,
            )
    """

    _impl: _ServiceImpl[RequestHeaders, ResponseHeaders]

    def __init__(self):
        self._impl = _ServiceImpl[RequestHeaders, ResponseHeaders]()

    def add_method(
        self,
        method: Method[Request, Response],
        impl: Union[
            Callable[[Request], Response],
            Callable[[Request, RequestHeaders], Response],
            Callable[[Request, RequestHeaders, ResponseHeaders], Response],
        ],
    ) -> None:
        self._impl.add_method(method, impl)

    def handle_request(
        self,
        req_body: str,
        req_headers: RequestHeaders,
        res_headers: ResponseHeaders,
    ) -> RawServiceResponse:
        return self._impl.handle_request(req_body, req_headers, res_headers)


class ServiceAsync(Generic[RequestHeaders, ResponseHeaders]):
    _impl: _ServiceImpl[RequestHeaders, ResponseHeaders]

    def __init__(self):
        self._impl = _ServiceImpl[RequestHeaders, ResponseHeaders]()

    def add_method(
        self,
        method: Method[Request, Response],
        impl: Union[
            # Sync
            Callable[[Request], Response],
            Callable[[Request, RequestHeaders], Response],
            Callable[[Request, RequestHeaders, ResponseHeaders], Response],
            # Async
            Callable[[Request], Awaitable[Response]],
            Callable[[Request, RequestHeaders], Awaitable[Response]],
            Callable[[Request, RequestHeaders, ResponseHeaders], Awaitable[Response]],
        ],
    ) -> None:
        self._impl.add_method(method, impl)

    async def handle_request(
        self,
        req_body: str,
        req_headers: RequestHeaders,
        res_headers: ResponseHeaders,
    ) -> RawServiceResponse:
        return await self._impl.handle_request_async(req_body, req_headers, res_headers)



import typing as _typing

from soia._impl.keyed_items import KeyedItems
from soia._impl.method import Method
from soia._impl.serializer import Serializer
from soia._impl.serializers import (
    array_serializer,
    optional_serializer,
    primitive_serializer,
)
from soia._impl.service import RawServiceResponse, Service, ServiceAsync
from soia._impl.service_client import ServiceClient
from soia._impl.timestamp import Timestamp

_: _typing.Final[_typing.Any] = None

__all__ = [
    "_",
    "KeyedItems",
    "Method",
    "RawServiceResponse",
    "Serializer",
    "Service",
    "ServiceAsync",
    "ServiceClient",
    "Timestamp",
    "array_serializer",
    "optional_serializer",
    "primitive_serializer",
]



import dataclasses
import unittest
from typing import Any

from soia._module_initializer import init_module
from soia.reflection import TypeDescriptor

from soia import KeyedItems, Method, Timestamp, _spec


class ModuleInitializerTestCase(unittest.TestCase):
    def init_test_module(self) -> dict[str, Any]:
        globals: dict[str, Any] = {}
        init_module(
            records=(
                _spec.Struct(
                    id="my/module.soia:Point",
                    fields=(
                        _spec.Field(
                            name="x",
                            number=0,
                            type=_spec.PrimitiveType.FLOAT32,
                        ),
                        _spec.Field(
                            name="y",
                            number=2,
                            type=_spec.PrimitiveType.FLOAT32,
                        ),
                    ),
                    removed_numbers=(1,),
                ),
                _spec.Struct(
                    id="my/module.soia:Segment",
                    fields=(
                        _spec.Field(
                            name="a",
                            number=0,
                            type="my/module.soia:Point",
                            has_mutable_getter=True,
                        ),
                        _spec.Field(
                            name="bb",
                            _attribute="b",
                            number=1,
                            type="my/module.soia:Point",
                            has_mutable_getter=True,
                        ),
                        _spec.Field(
                            name="c",
                            number=2,
                            type=_spec.OptionalType("my/module.soia:Point"),
                            has_mutable_getter=True,
                        ),
                    ),
                ),
                _spec.Struct(
                    id="my/module.soia:Shape",
                    fields=(
                        _spec.Field(
                            name="points",
                            number=0,
                            type=_spec.ArrayType("my/module.soia:Point"),
                            has_mutable_getter=True,
                        ),
                    ),
                ),
                _spec.Struct(
                    id="my/module.soia:Primitives",
                    fields=(
                        _spec.Field(
                            name="bool",
                            number=0,
                            type=_spec.PrimitiveType.BOOL,
                        ),
                        _spec.Field(
                            name="bytes",
                            number=1,
                            type=_spec.PrimitiveType.BYTES,
                        ),
                        _spec.Field(
                            name="f32",
                            number=2,
                            type=_spec.PrimitiveType.FLOAT32,
                        ),
                        _spec.Field(
                            name="f64",
                            number=3,
                            type=_spec.PrimitiveType.FLOAT64,
                        ),
                        _spec.Field(
                            name="i32",
                            number=4,
                            type=_spec.PrimitiveType.INT32,
                        ),
                        _spec.Field(
                            name="i64",
                            number=5,
                            type=_spec.PrimitiveType.INT64,
                        ),
                        _spec.Field(
                            name="u64",
                            number=6,
                            type=_spec.PrimitiveType.UINT64,
                        ),
                        _spec.Field(
                            name="s",
                            number=7,
                            type=_spec.PrimitiveType.STRING,
                        ),
                        _spec.Field(
                            name="t",
                            number=8,
                            type=_spec.PrimitiveType.TIMESTAMP,
                        ),
                    ),
                ),
                _spec.Struct(
                    id="my/module.soia:After",
                    fields=(
                        _spec.Field(
                            name="points",
                            number=0,
                            type=_spec.ArrayType("my/module.soia:Point"),
                            has_mutable_getter=True,
                        ),
                    ),
                ),
                _spec.Enum(
                    id="my/module.soia:PrimaryColor",
                    constant_fields=(
                        _spec.ConstantField(
                            name="RED",
                            number=10,
                        ),
                        _spec.ConstantField(
                            name="GREEN",
                            number=20,
                        ),
                        _spec.ConstantField(
                            name="BLUE",
                            number=30,
                        ),
                    ),
                ),
                _spec.Enum(
                    id="my/module.soia:Status",
                    constant_fields=(
                        _spec.ConstantField(
                            name="OK",
                            number=1,
                        ),
                    ),
                    value_fields=(
                        _spec.ValueField(
                            name="error",
                            number=2,
                            type=_spec.PrimitiveType.STRING,
                        ),
                    ),
                    removed_numbers=(1, 4),
                ),
                _spec.Enum(
                    id="my/module.soia:JsonValue",
                    constant_fields=(
                        _spec.ConstantField(
                            name="NULL",
                            number=1,
                        ),
                    ),
                    value_fields=(
                        _spec.ValueField(
                            name="bool",
                            number=2,
                            type=_spec.PrimitiveType.BOOL,
                        ),
                        _spec.ValueField(
                            name="number",
                            number=3,
                            type=_spec.PrimitiveType.FLOAT64,
                        ),
                        _spec.ValueField(
                            name="string",
                            number=4,
                            type=_spec.PrimitiveType.STRING,
                        ),
                        _spec.ValueField(
                            name="array",
                            number=5,
                            type=_spec.ArrayType("my/module.soia:JsonValue"),
                        ),
                        _spec.ValueField(
                            name="object",
                            number=6,
                            type="my/module.soia:JsonValue.Object",
                        ),
                    ),
                    removed_numbers=(
                        100,
                        101,
                    ),
                ),
                _spec.Struct(
                    id="my/module.soia:JsonValue.Object",
                    fields=(
                        _spec.Field(
                            name="entries",
                            number=0,
                            type=_spec.ArrayType(
                                item="my/module.soia:JsonValue.ObjectEntry",
                                key_attributes=("name",),
                            ),
                        ),
                    ),
                ),
                _spec.Struct(
                    id="my/module.soia:JsonValue.ObjectEntry",
                    fields=(
                        _spec.Field(
                            name="name",
                            number=0,
                            type=_spec.PrimitiveType.STRING,
                        ),
                        _spec.Field(
                            name="value",
                            number=1,
                            type="my/module.soia:JsonValue",
                        ),
                    ),
                ),
                _spec.Struct(
                    id="my/module.soia:Parent",
                    fields=(),
                ),
                _spec.Enum(
                    id="my/module.soia:Parent.NestedEnum",
                ),
                _spec.Struct(
                    id="my/module.soia:Stuff",
                    fields=(
                        _spec.Field(
                            name="enum_wrappers",
                            number=0,
                            type=_spec.ArrayType(
                                item="my/module.soia:EnumWrapper",
                                key_attributes=(
                                    "status",
                                    "kind",
                                ),
                            ),
                        ),
                    ),
                ),
                _spec.Struct(
                    id="my/module.soia:EnumWrapper",
                    fields=(
                        _spec.Field(
                            name="status",
                            number=0,
                            type="my/module.soia:Status",
                        ),
                    ),
                ),
                _spec.Struct(
                    id="my/module.soia:Stuff.Overrides",
                    _class_name="NameOverrides",
                    _class_qualname="Stuff.NameOverrides",
                    fields=(
                        _spec.Field(
                            name="x",
                            _attribute="y",
                            number=0,
                            type=_spec.PrimitiveType.INT32,
                        ),
                    ),
                ),
                _spec.Struct(
                    id="my/module.soia:RecOuter",
                    fields=(
                        _spec.Field(
                            name="r",
                            number=0,
                            type="my/module.soia:RecOuter.RecInner",
                        ),
                    ),
                ),
                _spec.Struct(
                    id="my/module.soia:RecOuter.RecInner",
                    fields=(
                        _spec.Field(
                            name="r",
                            number=0,
                            type="my/module.soia:RecOuter",
                        ),
                    ),
                ),
                _spec.Struct(
                    id="my/module.soia:Rec",
                    fields=(
                        _spec.Field(
                            name="r",
                            number=0,
                            type="my/module.soia:Rec",
                        ),
                        _spec.Field(
                            name="x",
                            number=1,
                            type=_spec.PrimitiveType.INT32,
                        ),
                    ),
                ),
                _spec.Struct(
                    id="my/module.soia:Foobar",
                    fields=(
                        _spec.Field(
                            name="a",
                            number=1,
                            type=_spec.PrimitiveType.INT32,
                        ),
                        _spec.Field(
                            name="b",
                            number=3,
                            type=_spec.PrimitiveType.INT32,
                        ),
                        _spec.Field(
                            name="point",
                            number=4,
                            type="my/module.soia:Point",
                        ),
                    ),
                    removed_numbers=(0, 2),
                ),
            ),
            methods=(
                _spec.Method(
                    name="FirstMethod",
                    number=-300,
                    request_type="my/module.soia:Point",
                    response_type="my/module.soia:Shape",
                ),
                _spec.Method(
                    name="SecondMethod",
                    number=-301,
                    request_type="my/module.soia:Point",
                    response_type="my/module.soia:Shape",
                    _var_name="MethodVar",
                ),
            ),
            constants=(
                _spec.Constant(
                    name="C",
                    type="my/module.soia:Point",
                    json_code="[1.5, 0, 2.5]",
                ),
            ),
            globals=globals,
            record_id_to_adapter={},
        )
        return globals

    def test_struct_getters(self):
        point_cls = self.init_test_module()["Point"]
        point = point_cls(x=1.5, y=2.5)
        self.assertEqual(point.x, 1.5)
        self.assertEqual(point.y, 2.5)

    def test_whole_static_factory_method(self):
        point_cls = self.init_test_module()["Point"]
        point = point_cls.whole(x=1.5, y=2.5)
        self.assertEqual(point.x, 1.5)
        self.assertEqual(point.y, 2.5)

    def test_to_mutable(self):
        point_cls = self.init_test_module()["Point"]
        point = point_cls(x=1.5, y=2.5)
        mutable = point.to_mutable()
        mutable.x = 4.0
        point = mutable.to_frozen()
        self.assertEqual(point.x, 4.0)
        self.assertEqual(point.y, 2.5)
        self.assertIs(point.to_frozen(), point)

    def test_struct_eq(self):
        point_cls = self.init_test_module()["Point"]
        a = point_cls(x=1.5, y=2.5)
        b = point_cls(x=1.5, y=2.5)
        c = point_cls(x=1.5, y=3.0)
        self.assertEqual(a, b)
        self.assertEqual(hash(a), hash(b))
        self.assertNotEqual(a, c)
        self.assertNotEqual(a, "foo")
        self.assertEqual(point_cls(), point_cls(x=0.0, y=0.0))

    def test_or_mutable(self):
        point_cls = self.init_test_module()["Point"]
        point_cls.OrMutable

    def test_primitives_default_values(self):
        primitives_cls = self.init_test_module()["Primitives"]
        a = primitives_cls(
            bool=False,
            bytes=b"",
            f32=0.0,
            f64=0,
            i32=0,
            i64=0,
            u64=0,
            t=Timestamp.EPOCH,
        )
        b = primitives_cls()
        self.assertEqual(a, b)
        self.assertEqual(hash(a), hash(b))

    def test_primitives_to_json(self):
        primitives_cls = self.init_test_module()["Primitives"]
        serializer = primitives_cls.SERIALIZER
        p = primitives_cls(
            bool=True,
            bytes=b"a",
            f32=3.14,
            f64=3.14,
            i32=1,
            i64=2,
            u64=3,
            t=Timestamp.from_unix_millis(4),
        )
        self.assertEqual(serializer.to_json(p), [1, "61", 3.14, 3.14, 1, 2, 3, "", 4])

    def test_primitives_from_json(self):
        primitives_cls = self.init_test_module()["Primitives"]
        serializer = primitives_cls.SERIALIZER
        json = [0] * 100
        self.assertEqual(serializer.from_json(json), primitives_cls.DEFAULT)
        self.assertEqual(
            serializer.from_json([1, "61", 3.14, 3.14, 1, 2, 3, "", 4]),
            primitives_cls(
                bool=True,
                bytes=b"a",
                f32=3.14,
                f64=3.14,
                i32=1,
                i64=2,
                u64=3,
                t=Timestamp.from_unix_millis(4),
            ),
        )

    def test_primitives_repr(self):
        primitives_cls = self.init_test_module()["Primitives"]
        p = primitives_cls(
            bool=True,
            bytes=b"a",
            f32=3.14,
            f64=3.14,
            i32=1,
            i64=2,
            u64=3,
            t=Timestamp.from_unix_millis(4),
        )
        self.assertEqual(
            str(p),
            "Primitives(\n  bool=True,\n  bytes=b'a',\n  f32=3.14,\n  f64=3.14,\n  i32=1,\n  i64=2,\n  u64=3,\n  t=Timestamp(\n    unix_millis=4,\n    _formatted='1970-01-01T00:00:00.004000Z',\n  ),\n)",
        )

    def test_from_json_converts_between_ints_and_floats(self):
        primitives_cls = self.init_test_module()["Primitives"]
        serializer = primitives_cls.SERIALIZER
        p = serializer.from_json([0, 0, 3])
        self.assertEqual(p.f32, 3.0)
        self.assertIsInstance(p.f32, float)
        p = serializer.from_json({"i32": 1.2})
        self.assertEqual(p.i32, 1)
        self.assertIsInstance(p.i32, int)

    def test_cannot_mutate_frozen_class(self):
        point_cls = self.init_test_module()["Point"]
        point = point_cls(x=1.5, y=2.5)
        try:
            point.x = 3.5
            self.fail("expected to raise FrozenInstanceError")
        except dataclasses.FrozenInstanceError:
            pass
        self.assertEqual(point.x, 1.5)

    def test_point_to_dense_json(self):
        point_cls = self.init_test_module()["Point"]
        serializer = point_cls.SERIALIZER
        point = point_cls(x=1.5, y=2.5)
        self.assertEqual(serializer.to_json(point), [1.5, 0, 2.5])
        self.assertEqual(serializer.to_json(point, readable=False), [1.5, 0, 2.5])
        self.assertEqual(serializer.to_json_code(point), "[1.5,0,2.5]")
        point = point_cls(x=1.5, y=0.0)
        self.assertEqual(serializer.to_json(point), [1.5])

    def test_point_to_readable_json(self):
        point_cls = self.init_test_module()["Point"]
        point = point_cls(x=1.5, y=2.5)
        json = point_cls.SERIALIZER.to_json(point, readable=True)
        self.assertEqual(json, {"x": 1.5, "y": 2.5})
        json_code = point_cls.SERIALIZER.to_json_code(point, readable=True)
        self.assertEqual(json_code, '{\n  "x": 1.5,\n  "y": 2.5\n}')

    def test_point_from_dense_json(self):
        point_cls = self.init_test_module()["Point"]
        serializer = point_cls.SERIALIZER
        self.assertEqual(serializer.from_json([1.5, 0, 2.5]), point_cls(x=1.5, y=2.5))
        self.assertEqual(serializer.from_json([1.5]), point_cls(x=1.5))
        self.assertEqual(serializer.from_json([0.0]), point_cls.DEFAULT)
        self.assertEqual(serializer.from_json(0), point_cls.DEFAULT)

    def test_point_from_readable_json(self):
        point_cls = self.init_test_module()["Point"]
        point = point_cls.SERIALIZER.from_json({"x": 1.5, "y": 2.5})
        self.assertEqual(point, point_cls(x=1.5, y=2.5))
        point = point_cls.SERIALIZER.from_json_code('{"x":1.5,"y":2.5}')
        self.assertEqual(point, point_cls(x=1.5, y=2.5))
        point = point_cls.SERIALIZER.from_json_code('{"x":1.5,"y":2.5,"z":[]}')
        self.assertEqual(point, point_cls(x=1.5, y=2.5))
        point = point_cls.SERIALIZER.from_json_code('{"x":1,"y":2}')
        self.assertEqual(point.x, 1.0)
        self.assertIsInstance(point.x, float)
        self.assertEqual(point._array_len, 3)

    def test_point_with_unrecognized_and_removed_fields(self):
        point_cls = self.init_test_module()["Point"]
        serializer = point_cls.SERIALIZER
        point = serializer.from_json([1.5, 1, 2.5, True])
        self.assertEqual(point, point_cls(x=1.5, y=2.5))
        self.assertEqual(serializer.to_json(point), [1.5, 0, 2.5, True])
        point = point.to_mutable().to_frozen()
        self.assertEqual(serializer.to_json(point), [1.5, 0, 2.5, True])

    def test_struct_to_dense_json_with_removed_fields(self):
        test_module = self.init_test_module()
        foobar_cls = test_module["Foobar"]
        point_cls = test_module["Point"]
        serializer = foobar_cls.SERIALIZER
        foobar = foobar_cls()
        self.assertEqual(serializer.to_json_code(foobar), "[]")
        self.assertEqual(serializer.from_json_code("[]"), foobar)
        foobar = foobar_cls(a=3)
        self.assertEqual(serializer.to_json_code(foobar), "[0,3]")
        self.assertEqual(serializer.from_json_code("[0,3]"), foobar)
        self.assertEqual(serializer.from_json_code("[0,3.1]"), foobar)
        foobar = foobar_cls(b=3, point=point_cls.DEFAULT)
        self.assertEqual(serializer.to_json_code(foobar), "[0,0,0,3]")
        self.assertEqual(serializer.from_json_code("[0,0,0,3]"), foobar)
        self.assertEqual(serializer.from_json_code("[0,0,0,3.1]"), foobar)
        foobar = foobar_cls(point=point_cls(x=2))
        self.assertEqual(serializer.to_json_code(foobar), "[0,0,0,0,[2.0]]")
        self.assertEqual(serializer.from_json_code("[0,0,0,0,[2.0]]"), foobar)

    def test_recursive_struct(self):
        rec_cls = self.init_test_module()["Rec"]
        r = rec_cls(r=rec_cls(r=rec_cls.DEFAULT, x=1))
        serializer = rec_cls.SERIALIZER
        self.assertEqual(serializer.to_json_code(r), "[[[],1]]")
        self.assertEqual(serializer.from_json_code("[[[],1]]"), r)
        self.assertEqual(str(r), "Rec(\n  r=Rec(x=1),\n)")

    def test_struct_ctor_accepts_mutable_struct(self):
        module = self.init_test_module()
        segment_cls = module["Segment"]
        point_cls = module["Point"]
        segment = segment_cls(
            a=point_cls(x=1.0, y=2.0).to_mutable(),
            b=point_cls(x=3.0, y=4.0),
        )
        self.assertEqual(
            segment,
            segment_cls(
                a=point_cls(x=1.0, y=2.0),
                b=point_cls(x=3.0, y=4.0),
                c=None,
            ),
        )

    def test_struct_ctor_checks_type_of_struct_param(self):
        module = self.init_test_module()
        segment_cls = module["Segment"]
        try:
            segment_cls(
                # Should be a Point
                a=segment_cls.DEFAULT,
            )
            self.fail("Expected to fail")
        except Exception as e:
            self.assertIn("Point", str(e))

    def test_struct_ctor_raises_error_if_unknown_arg(self):
        module = self.init_test_module()
        segment_cls = module["Segment"]
        try:
            segment_cls(foo=4)
            self.fail("Expected to fail")
        except Exception:
            pass

    def test_to_frozen_checks_type_of_struct_field(self):
        module = self.init_test_module()
        segment_cls = module["Segment"]
        mutable = segment_cls.Mutable()
        mutable.a = segment_cls.DEFAULT  # Should be a Point
        try:
            mutable.to_frozen()
            self.fail("Expected to fail")
        except Exception as e:
            self.assertIn("Point", str(e))

    def test_struct_ctor_accepts_mutable_list(self):
        module = self.init_test_module()
        shape_cls = module["Shape"]
        point_cls = module["Point"]
        shape = shape_cls(
            points=[
                point_cls(x=1.0, y=2.0).to_mutable(),
                point_cls(x=3.0, y=4.0),
            ],
        )
        self.assertEqual(
            shape,
            shape_cls(
                points=(
                    point_cls(x=1.0, y=2.0),
                    point_cls(x=3.0, y=4.0),
                ),
            ),
        )

    def test_listuple_not_copied(self):
        module = self.init_test_module()
        shape_cls = module["Shape"]
        point_cls = module["Point"]
        shape = shape_cls(
            points=[
                point_cls(x=1.0, y=2.0),
                point_cls(x=3.0, y=4.0),
            ],
        )
        other_shape = shape.to_mutable().to_frozen()
        self.assertIsNot(other_shape, shape)
        self.assertIs(other_shape.points, shape.points)
        self.assertIsNot(other_shape.points.__class__, tuple)

    def test_single_empty_listuple_instance(self):
        module = self.init_test_module()
        shape_cls = module["Shape"]
        shape = shape_cls(
            points=[],
        )
        self.assertIs(shape.points, shape_cls(points=[]).points)
        self.assertIs(shape.points, shape.to_mutable().to_frozen().points)
        self.assertIsNot(shape.points, ())

    def test_optional(self):
        module = self.init_test_module()
        segment_cls = module["Segment"]
        point_cls = module["Point"]
        segment = segment_cls(
            c=point_cls.Mutable(x=1.0, y=2.0),
        )
        other_segment = segment.to_mutable().to_frozen()
        self.assertEqual(
            other_segment,
            segment_cls(
                c=point_cls(x=1.0, y=2.0),
            ),
        )

    def test_enum_unknown_constant(self):
        module = self.init_test_module()
        primary_color_cls = module["PrimaryColor"]
        unknown = primary_color_cls.UNKNOWN
        self.assertEqual(unknown.kind, "?")
        self.assertEqual(unknown.value, None)
        self.assertIs(unknown.union, unknown)
        serializer = primary_color_cls.SERIALIZER
        self.assertEqual(serializer.to_json(unknown), 0)
        self.assertEqual(serializer.to_json(unknown, readable=True), "?")

    def test_enum_user_defined_constant(self):
        module = self.init_test_module()
        primary_color_cls = module["PrimaryColor"]
        red = primary_color_cls.RED
        self.assertEqual(red.kind, "RED")
        self.assertEqual(red.value, None)
        self.assertIs(red.union, red)
        serializer = primary_color_cls.SERIALIZER
        self.assertEqual(serializer.to_json(red), 10)
        self.assertEqual(serializer.to_json(red, readable=True), "RED")

    def test_enum_wrap(self):
        module = self.init_test_module()
        status_cls = module["Status"]
        error = status_cls.wrap_error("An error occurred")
        self.assertEqual(error.kind, "error")
        self.assertEqual(error.value, "An error occurred")
        self.assertIs(error.union, error)
        serializer = status_cls.SERIALIZER
        self.assertEqual(serializer.to_json(error), [2, "An error occurred"])
        self.assertEqual(
            serializer.to_json(error, readable=True),
            {"kind": "error", "value": "An error occurred"},
        )

    def test_enum_wrap_around_mutable_struct(self):
        module = self.init_test_module()
        json_value_cls = module["JsonValue"]
        json_object_cls = json_value_cls.Object
        json_object = json_value_cls.wrap_object(json_object_cls().to_mutable())
        self.assertEqual(json_object.kind, "object")
        self.assertEqual(json_object.value, json_object_cls.DEFAULT)
        self.assertEqual(
            json_object, json_value_cls.wrap_object(json_object_cls.DEFAULT)
        )

    def test_enum_to_json(self):
        module = self.init_test_module()
        status_cls = module["Status"]
        serializer = status_cls.SERIALIZER
        self.assertEqual(serializer.to_json(status_cls.UNKNOWN), 0)
        self.assertEqual(serializer.to_json(status_cls.UNKNOWN, readable=True), "?")
        self.assertEqual(serializer.to_json(status_cls.OK), 1)
        self.assertEqual(serializer.to_json(status_cls.OK, readable=True), "OK")
        self.assertEqual(serializer.to_json(status_cls.OK, readable=False), 1)
        self.assertEqual(serializer.to_json(status_cls.wrap_error("E")), [2, "E"])
        self.assertEqual(
            serializer.to_json(status_cls.wrap_error("E"), readable=True),
            {"kind": "error", "value": "E"},
        )

    def test_enum_from_json(self):
        module = self.init_test_module()
        status_cls = module["Status"]
        serializer = status_cls.SERIALIZER
        self.assertEqual(serializer.from_json(0), status_cls.UNKNOWN)
        self.assertEqual(serializer.from_json("?"), status_cls.UNKNOWN)
        self.assertEqual(serializer.from_json(1), status_cls.OK)
        self.assertEqual(serializer.from_json("OK"), status_cls.OK)
        self.assertEqual(serializer.from_json([2, "E"]), status_cls.wrap_error("E"))
        self.assertEqual(
            serializer.from_json({"kind": "error", "value": "E"}),
            status_cls.wrap_error("E"),
        )

    def test_complex_enum_from_json(self):
        module = self.init_test_module()
        json_value_cls = module["JsonValue"]
        serializer = json_value_cls.SERIALIZER
        json_value = serializer.from_json(
            [
                5,
                [
                    0,
                    1,
                    [2, True],
                    [3, 3.14],
                    [4, "foo"],
                    [
                        6,
                        [["a", 0], ["b", 0]],
                    ],
                    [5, [[5, []]]],
                ],
            ]
        )
        self.assertEqual(
            json_value,
            json_value_cls.wrap_array(
                [
                    json_value_cls.UNKNOWN,
                    json_value_cls.NULL,
                    json_value_cls.wrap_bool(True),
                    json_value_cls.wrap_number(3.14),
                    json_value_cls.wrap_string("foo"),
                    json_value_cls.wrap_object(
                        json_value_cls.Object(
                            entries=[
                                json_value_cls.ObjectEntry(),
                                json_value_cls.ObjectEntry.DEFAULT,
                            ],
                        )
                    ),
                    json_value_cls.wrap_array(
                        [
                            json_value_cls.wrap_array([]),
                        ]
                    ),
                ]
            ),
        )
        self.assertEqual(
            serializer.from_json(
                {
                    "kind": "array",
                    "value": [
                        "?",
                        "NULL",
                        {"kind": "bool", "value": True},
                        {"kind": "number", "value": 3.14},
                        {"kind": "string", "value": "foo"},
                        {"kind": "object", "value": {"entries": [{}, {}]}},
                        {"kind": "array", "value": [{"kind": "array", "value": []}]},
                    ],
                }
            ),
            json_value,
        )

    def test_struct_with_enum_field(self):
        json_value_cls = self.init_test_module()["JsonValue"]
        json_object_entry_cls = json_value_cls.ObjectEntry
        self.assertEqual(json_object_entry_cls.DEFAULT.value, json_value_cls.UNKNOWN)
        self.assertEqual(json_object_entry_cls().value, json_value_cls.UNKNOWN)

    def test_enum_with_unrecognized_and_removed_fields(self):
        json_value_cls = self.init_test_module()["JsonValue"]
        serializer = json_value_cls.SERIALIZER
        json_value = serializer.from_json(100)  # removed
        self.assertEqual(json_value, json_value_cls.UNKNOWN)
        self.assertEqual(serializer.to_json(json_value), 0)
        json_value = serializer.from_json([101, True])  # removed
        self.assertEqual(json_value, json_value_cls.UNKNOWN)
        self.assertEqual(serializer.to_json(json_value), 0)
        json_value = serializer.from_json(102)  # unrecognized
        self.assertEqual(json_value, json_value_cls.UNKNOWN)
        self.assertEqual(serializer.to_json(json_value), 102)
        json_value = serializer.from_json([102, True])  # unrecognized
        self.assertEqual(json_value, json_value_cls.UNKNOWN)
        self.assertEqual(serializer.to_json(json_value), [102, True])

    def test_class_name(self):
        module = self.init_test_module()
        shape_cls = module["Shape"]
        json_value_cls = module["JsonValue"]
        json_object_cls = json_value_cls.Object
        self.assertEqual(shape_cls.__name__, "Shape")
        self.assertEqual(shape_cls.__qualname__, "Shape")
        self.assertEqual(json_value_cls.__name__, "JsonValue")
        self.assertEqual(json_value_cls.__qualname__, "JsonValue")
        self.assertEqual(json_object_cls.__name__, "Object")
        self.assertEqual(json_object_cls.__qualname__, "JsonValue.Object")

    def test_struct_repr(self):
        module = self.init_test_module()
        point_cls = module["Point"]
        self.assertEqual(
            repr(point_cls(x=1.5)),
            "Point(x=1.5)",
        )
        self.assertEqual(
            repr(point_cls(x=1.5).to_mutable()),
            "\n".join(
                [
                    "Point.Mutable(",
                    "  x=1.5,",
                    "  y=0.0,",
                    ")",
                ]
            ),
        )
        self.assertEqual(
            repr(point_cls(x=1.5, y=2.5)),
            "\n".join(
                [
                    "Point(",
                    "  x=1.5,",
                    "  y=2.5,",
                    ")",
                ]
            ),
        )
        self.assertEqual(
            repr(point_cls()),
            "Point()",
        )
        self.assertEqual(
            repr(point_cls.DEFAULT),
            "Point.DEFAULT",
        )
        self.assertEqual(
            repr(point_cls.DEFAULT.to_mutable()),
            "\n".join(
                [
                    "Point.Mutable(",
                    "  x=0.0,",
                    "  y=0.0,",
                    ")",
                ]
            ),
        )
        shape_cls = module["Shape"]
        self.assertEqual(
            repr(shape_cls(points=[])),
            "Shape()",
        )
        self.assertEqual(
            repr(shape_cls(points=[]).to_mutable()),
            "Shape.Mutable(points=[])",
        )
        self.assertEqual(
            repr(shape_cls(points=[point_cls(x=1.5)])),
            "\n".join(
                [
                    "Shape(",
                    "  points=[",
                    "    Point(x=1.5),",
                    "  ],",
                    ")",
                ]
            ),
        )
        self.assertEqual(
            repr(
                shape_cls(
                    points=[
                        point_cls(x=1.5),
                        point_cls(y=2.5),
                    ],
                )
            ),
            "\n".join(
                [
                    "Shape(",
                    "  points=[",
                    "    Point(x=1.5),",
                    "    Point(y=2.5),",
                    "  ],",
                    ")",
                ]
            ),
        )
        self.assertEqual(
            repr(
                shape_cls.Mutable(
                    points=[
                        point_cls(x=1.5),
                        point_cls(y=2.5).to_mutable(),
                    ]
                )
            ),
            "\n".join(
                [
                    "Shape.Mutable(",
                    "  points=[",
                    "    Point(x=1.5),",
                    "    Point.Mutable(",
                    "      x=0.0,",
                    "      y=2.5,",
                    "    ),",
                    "  ],",
                    ")",
                ]
            ),
        )

    def test_enum_constant_repr(self):
        module = self.init_test_module()
        primary_color_cls = module["PrimaryColor"]
        parent_cls = module["Parent"]
        nested_enum_cls = parent_cls.NestedEnum
        self.assertEqual(repr(primary_color_cls.UNKNOWN), "PrimaryColor.UNKNOWN")
        self.assertEqual(repr(primary_color_cls.RED), "PrimaryColor.RED")
        self.assertEqual(repr(nested_enum_cls.UNKNOWN), "Parent.NestedEnum.UNKNOWN")

    def test_enum_value_repr(self):
        module = self.init_test_module()
        status_cls = module["Status"]
        json_value_cls = module["JsonValue"]
        json_object_cls = json_value_cls.Object
        self.assertEqual(
            repr(status_cls.wrap_error("An error")),
            "Status.wrap_error('An error')",
        )
        self.assertEqual(
            repr(status_cls.wrap_error("multiple\nlines\n")),
            "\n".join(
                [
                    "Status.wrap_error(",
                    "  '\\n'.join([",
                    "    'multiple',",
                    "    'lines',",
                    "    '',",
                    "  ])",
                    ")",
                ]
            ),
        )
        self.assertEqual(
            repr(json_value_cls.wrap_object(json_object_cls().DEFAULT)),
            "JsonValue.wrap_object(JsonValue.Object.DEFAULT)",
        )
        self.assertEqual(
            repr(json_value_cls.wrap_object(json_object_cls())),
            "\n".join(
                [
                    "JsonValue.wrap_object(",
                    "  JsonValue.Object()",
                    ")",
                ]
            ),
        )

    def test_find_in_keyed_items(self):
        json_value_cls = self.init_test_module()["JsonValue"]
        object_cls = json_value_cls.Object
        entry_cls = json_value_cls.ObjectEntry
        json_object = object_cls(
            entries=[
                entry_cls(
                    name="foo",
                    value=json_value_cls.wrap_string("value of foo"),
                ),
                entry_cls(
                    name="bar",
                    value=json_value_cls.wrap_string("value of bar #0"),
                ),
                entry_cls(
                    name="foobar",
                    value=json_value_cls.wrap_string("value of foobar"),
                ),
                entry_cls(
                    name="bar",
                    value=json_value_cls.wrap_string("value of bar #1"),
                ),
            ]
        )
        entries = json_object.entries
        self.assertIsInstance(entries, KeyedItems)
        self.assertIs(entries.find("foo"), entries[0])
        self.assertIs(entries.find("zoo"), None)
        self.assertIs(entries.find("bar"), entries[3])
        self.assertIs(entries.find_or_default("foo"), entries[0])
        self.assertIs(entries.find_or_default("zoo"), entry_cls.DEFAULT)

    def test_find_in_keyed_items_with_complex_path(self):
        module = self.init_test_module()
        stuff_cls = module["Stuff"]
        enum_wrapper_cls = module["EnumWrapper"]
        status_cls = module["Status"]
        stuff = stuff_cls(
            enum_wrappers=[
                enum_wrapper_cls(
                    status=status_cls.OK,
                ),
                enum_wrapper_cls(
                    status=status_cls.wrap_error("an error"),
                ),
                enum_wrapper_cls(
                    status=status_cls.OK,
                ),
            ]
        )
        enum_wrappers = stuff.enum_wrappers
        self.assertIsInstance(enum_wrappers, KeyedItems)
        self.assertIs(enum_wrappers.find("OK"), enum_wrappers[2])
        self.assertIs(enum_wrappers.find("error"), enum_wrappers[1])
        self.assertIs(enum_wrappers.find_or_default("?"), enum_wrapper_cls.DEFAULT)
        enum_wrappers = stuff.to_mutable().to_frozen().enum_wrappers
        self.assertIs(enum_wrappers.find("error"), enum_wrappers[1])

    def test_name_overrides(self):
        name_overrides_cls = self.init_test_module()["Stuff"].NameOverrides
        name_overrides = name_overrides_cls(y=3)
        self.assertEqual(name_overrides.y, 3)
        self.assertEqual(name_overrides_cls.__name__, "NameOverrides")
        self.assertEqual(name_overrides_cls.__qualname__, "Stuff.NameOverrides")

    def test_mutable_getter_of_struct(self):
        module = self.init_test_module()
        segment_cls = module["Segment"]
        point_cls = module["Point"]
        segment = segment_cls(
            a=point_cls(x=1.0, y=2.0),
            b=point_cls(x=3.0, y=4.0),
        ).to_mutable()
        a = segment.mutable_a
        self.assertIsInstance(a, point_cls.Mutable)
        self.assertIs(segment.mutable_a, a)
        segment.a = "foo"
        try:
            segment.mutable_a()
            self.fail("Expected to fail")
        except TypeError as e:
            self.assertEqual(str(e), "expected: Point or Point.Mutable; found: str")

    def test_mutable_getter_of_array(self):
        module = self.init_test_module()
        shape_cls = module["Shape"]
        point_cls = module["Point"]
        shape = shape_cls(
            points=[
                point_cls(x=1.0, y=2.0),
                point_cls(x=3.0, y=4.0),
            ],
        ).to_mutable()
        points = shape.mutable_points
        self.assertIsInstance(points, list)
        self.assertIs(shape.mutable_points, points)

    def test_methods(self):
        module = self.init_test_module()
        first_method = module["FirstMethod"]
        self.assertEqual(
            first_method,
            Method(
                name="FirstMethod",
                number=-300,
                request_serializer=module["Point"].SERIALIZER,
                response_serializer=module["Shape"].SERIALIZER,
            ),
        )
        second_method = module["MethodVar"]
        self.assertEqual(
            second_method,
            Method(
                name="SecondMethod",
                number=-301,
                request_serializer=module["Point"].SERIALIZER,
                response_serializer=module["Shape"].SERIALIZER,
            ),
        )

    def test_constants(self):
        module = self.init_test_module()
        c = module["C"]
        Point = module["Point"]
        self.assertEqual(c, Point(x=1.5, y=2.5))

    def test_enum_type_descriptor(self):
        module = self.init_test_module()
        json_value_cls = module["JsonValue"]
        type_descriptor = json_value_cls.SERIALIZER.type_descriptor
        self.assertEqual(
            type_descriptor.as_json(),
            {
                "type": {"kind": "record", "value": "my/module.soia:JsonValue"},
                "records": [
                    {
                        "kind": "enum",
                        "id": "my/module.soia:JsonValue",
                        "fields": [
                            {
                                "name": "bool",
                                "type": {"kind": "primitive", "value": "bool"},
                                "number": 2,
                            },
                            {
                                "name": "number",
                                "type": {"kind": "primitive", "value": "float64"},
                                "number": 3,
                            },
                            {
                                "name": "string",
                                "type": {"kind": "primitive", "value": "string"},
                                "number": 4,
                            },
                            {
                                "name": "array",
                                "type": {
                                    "kind": "array",
                                    "value": {
                                        "item": {
                                            "kind": "record",
                                            "value": "my/module.soia:JsonValue",
                                        },
                                    },
                                },
                                "number": 5,
                            },
                            {
                                "name": "object",
                                "type": {
                                    "kind": "record",
                                    "value": "my/module.soia:JsonValue.Object",
                                },
                                "number": 6,
                            },
                        ],
                        "removed_fields": [100, 101],
                    },
                    {
                        "kind": "struct",
                        "id": "my/module.soia:JsonValue.Object",
                        "fields": [
                            {
                                "name": "entries",
                                "type": {
                                    "kind": "array",
                                    "value": {
                                        "item": {
                                            "kind": "record",
                                            "value": "my/module.soia:JsonValue.ObjectEntry",
                                        },
                                        "key_chain": ["name"],
                                    },
                                },
                                "number": 0,
                            }
                        ],
                    },
                    {
                        "kind": "struct",
                        "id": "my/module.soia:JsonValue.ObjectEntry",
                        "fields": [
                            {
                                "name": "name",
                                "type": {"kind": "primitive", "value": "string"},
                                "number": 0,
                            },
                            {
                                "name": "value",
                                "type": {
                                    "kind": "record",
                                    "value": "my/module.soia:JsonValue",
                                },
                                "number": 1,
                            },
                        ],
                    },
                ],
            },
        )
        self.assertEqual(
            str(TypeDescriptor.from_json(type_descriptor.as_json())),
            str(type_descriptor),
        )
        self.assertEqual(
            TypeDescriptor.from_json(type_descriptor.as_json()),
            type_descriptor,
        )

    def test_optional_type_descriptor(self):
        module = self.init_test_module()
        segment_cls = module["Segment"]
        type_descriptor = segment_cls.SERIALIZER.type_descriptor
        self.assertEqual(
            type_descriptor.as_json(),
            {
                "type": {"kind": "record", "value": "my/module.soia:Segment"},
                "records": [
                    {
                        "kind": "struct",
                        "id": "my/module.soia:Segment",
                        "fields": [
                            {
                                "name": "a",
                                "type": {
                                    "kind": "record",
                                    "value": "my/module.soia:Point",
                                },
                                "number": 0,
                            },
                            {
                                "name": "bb",
                                "type": {
                                    "kind": "record",
                                    "value": "my/module.soia:Point",
                                },
                                "number": 1,
                            },
                            {
                                "name": "c",
                                "type": {
                                    "kind": "optional",
                                    "value": {
                                        "kind": "record",
                                        "value": "my/module.soia:Point",
                                    },
                                },
                                "number": 2,
                            },
                        ],
                    },
                    {
                        "kind": "struct",
                        "id": "my/module.soia:Point",
                        "fields": [
                            {
                                "name": "x",
                                "type": {"kind": "primitive", "value": "float32"},
                                "number": 0,
                            },
                            {
                                "name": "y",
                                "type": {"kind": "primitive", "value": "float32"},
                                "number": 2,
                            },
                        ],
                        "removed_fields": [1],
                    },
                ],
            },
        )
        self.assertEqual(
            str(TypeDescriptor.from_json(type_descriptor.as_json())),
            str(type_descriptor),
        )
        self.assertEqual(
            TypeDescriptor.from_json(type_descriptor.as_json()),
            type_descriptor,
        )

    def test_primitive_type_descriptor(self):
        module = self.init_test_module()
        segment_cls = module["Primitives"]
        type_descriptor = segment_cls.SERIALIZER.type_descriptor
        self.assertEqual(
            type_descriptor.as_json(),
            {
                "type": {"kind": "record", "value": "my/module.soia:Primitives"},
                "records": [
                    {
                        "kind": "struct",
                        "id": "my/module.soia:Primitives",
                        "fields": [
                            {
                                "name": "bool",
                                "type": {"kind": "primitive", "value": "bool"},
                                "number": 0,
                            },
                            {
                                "name": "bytes",
                                "type": {"kind": "primitive", "value": "bytes"},
                                "number": 1,
                            },
                            {
                                "name": "f32",
                                "type": {"kind": "primitive", "value": "float32"},
                                "number": 2,
                            },
                            {
                                "name": "f64",
                                "type": {"kind": "primitive", "value": "float64"},
                                "number": 3,
                            },
                            {
                                "name": "i32",
                                "type": {"kind": "primitive", "value": "int32"},
                                "number": 4,
                            },
                            {
                                "name": "i64",
                                "type": {"kind": "primitive", "value": "int64"},
                                "number": 5,
                            },
                            {
                                "name": "u64",
                                "type": {"kind": "primitive", "value": "uint64"},
                                "number": 6,
                            },
                            {
                                "name": "s",
                                "type": {"kind": "primitive", "value": "string"},
                                "number": 7,
                            },
                            {
                                "name": "t",
                                "type": {"kind": "primitive", "value": "timestamp"},
                                "number": 8,
                            },
                        ],
                    }
                ],
            },
        )
        self.assertEqual(
            str(TypeDescriptor.from_json(type_descriptor.as_json())),
            str(type_descriptor),
        )
        self.assertEqual(
            TypeDescriptor.from_json(type_descriptor.as_json()),
            type_descriptor,
        )



import re
from dataclasses import dataclass
from typing import Any


@dataclass
class ReprResult:
    repr: str
    complex: bool

    @property
    def indented(self) -> str:
        if complex:
            return self.repr.replace("\n", "\n  ")
        else:
            return self.repr


def repr_impl(input: Any) -> ReprResult:
    if isinstance(input, list) or isinstance(input, tuple):
        if len(input) == 0:
            return ReprResult("[]", complex=False)
        elif len(input) == 1:
            only_item_repr = repr_impl(input[0])
            if only_item_repr.complex:
                return ReprResult(f"[\n  {only_item_repr.indented},\n]", complex=True)
            else:
                return ReprResult(f"[{only_item_repr.repr}]", complex=True)
        else:
            body = "".join(f"  {repr_impl(item).indented},\n" for item in input)
            return ReprResult(f"[\n{body}]", complex=True)
    elif isinstance(input, str):
        if "\n" in input:
            lines = input.split("\n")
            body = "".join(f"  {repr(line)},\n" for line in lines)
            return ReprResult(f"'\\n'.join([\n{body}])", complex=True)
        else:
            return ReprResult(repr(input), complex=False)
    else:
        r = repr(input)
        # Complex if the representation contains '[', '(' or '\n'.
        return ReprResult(r, complex=re.search(r"[\\[\\(\\\n]", r) is not None)



import dataclasses
import datetime
from typing import Any, Final, Union, cast, final, overload


@final
class Timestamp:
    __slots__ = ("unix_millis",)

    unix_millis: int

    def __init__(self, unix_millis: int, _formatted: str = ""):
        object.__setattr__(
            self,
            "unix_millis",
            round(min(max(unix_millis, -8640000000000000), 8640000000000000)),
        )

    @staticmethod
    def from_unix_millis(unix_millis: int) -> "Timestamp":
        return Timestamp(unix_millis=unix_millis)

    @staticmethod
    def from_unix_seconds(unix_seconds: float) -> "Timestamp":
        return Timestamp(unix_millis=round(unix_seconds * 1000))

    @staticmethod
    def from_datetime(dt: datetime.datetime) -> "Timestamp":
        return Timestamp.from_unix_seconds(dt.timestamp())

    @staticmethod
    def now() -> "Timestamp":
        return Timestamp.from_datetime(datetime.datetime.now(tz=datetime.timezone.utc))

    EPOCH: Final["Timestamp"] = cast("Timestamp", ...)
    MIN: Final["Timestamp"] = cast("Timestamp", ...)
    MAX: Final["Timestamp"] = cast("Timestamp", ...)

    @property
    def unix_seconds(self) -> float:
        return self.unix_millis / 1000.0

    def to_datetime_or_raise(self) -> datetime.datetime:
        return datetime.datetime.fromtimestamp(
            self.unix_seconds, tz=datetime.timezone.utc
        )

    def __add__(self, td: datetime.timedelta) -> "Timestamp":
        return Timestamp(
            unix_millis=self.unix_millis + round(td.total_seconds() * 1000)
        )

    @overload
    def __sub__(self, other: datetime.timedelta) -> "Timestamp": ...
    @overload
    def __sub__(self, other: "Timestamp") -> datetime.timedelta: ...

    def __sub__(
        self, other: Union["Timestamp", datetime.timedelta]
    ) -> Union["Timestamp", datetime.timedelta]:
        if isinstance(other, Timestamp):
            return datetime.timedelta(milliseconds=self.unix_millis - other.unix_millis)
        else:
            return self.__add__(-other)

    def __lt__(self, other: "Timestamp"):
        return self.unix_millis < other.unix_millis

    def __gt__(self, other: "Timestamp"):
        return self.unix_millis > other.unix_millis

    def __le__(self, other: "Timestamp"):
        return self.unix_millis <= other.unix_millis

    def __ge__(self, other: "Timestamp"):
        return self.unix_millis >= other.unix_millis

    def __eq__(self, other: Any):
        if isinstance(other, Timestamp):
            return other.unix_millis == self.unix_millis
        return NotImplemented

    def __hash__(self):
        return hash(("ts", self.unix_millis))

    def __repr__(self) -> str:
        iso = self._iso_format()
        if iso:
            return f"Timestamp(\n  unix_millis={self.unix_millis},\n  _formatted='{iso}',\n)"
        else:
            return f"Timestamp(unix_millis={self.unix_millis})"

    def __setattr__(self, name: str, value: Any):
        raise dataclasses.FrozenInstanceError(self.__class__.__qualname__)

    def __delattr__(self, name: str):
        raise dataclasses.FrozenInstanceError(self.__class__.__qualname__)

    def _trj(self) -> Any:
        """To readable JSON."""
        iso = self._iso_format()
        if iso:
            return {
                "unix_millis": self.unix_millis,
                "formatted": iso,
            }
        else:
            return {
                "unix_millis": self.unix_millis,
            }

    def _iso_format(self) -> str:
        try:
            dt = self.to_datetime_or_raise()
        except Exception:
            return ""
        if dt:
            ret = dt.isoformat()
            bad_suffix = "+00:00"
            if ret.endswith(bad_suffix):
                ret = ret[0 : -len(bad_suffix)] + "Z"
        return ret


# Use 'setattr' because we marked these class attributes as Final.
setattr(Timestamp, "EPOCH", Timestamp.from_unix_millis(0))
setattr(Timestamp, "MIN", Timestamp.from_unix_millis(-8640000000000000))
setattr(Timestamp, "MAX", Timestamp.from_unix_millis(8640000000000000))



from collections.abc import Callable
from dataclasses import dataclass
from typing import TypeVar
from weakref import WeakValueDictionary

from soia._impl.function_maker import Expr, ExprLike
from soia._impl.type_adapter import TypeAdapter

from soia import _spec, reflection

Other = TypeVar("Other")


def get_optional_adapter(other_adapter: TypeAdapter) -> TypeAdapter:
    return _other_adapter_to_optional_adapter.setdefault(
        other_adapter, _OptionalAdapter(other_adapter)
    )


@dataclass(frozen=True)
class _OptionalAdapter(TypeAdapter):
    __slots__ = ("other_adapter",)

    other_adapter: TypeAdapter

    def default_expr(self) -> ExprLike:
        return "None"

    def to_frozen_expr(self, arg_expr: ExprLike) -> Expr:
        return Expr.join(
            "(None if ",
            arg_expr,
            " is None else ",
            self.other_adapter.to_frozen_expr(arg_expr),
            ")",
        )

    def is_not_default_expr(self, arg_expr: ExprLike, attr_expr: ExprLike) -> Expr:
        return Expr.join(arg_expr, " is not None")

    def to_json_expr(self, in_expr: ExprLike, readable: bool) -> ExprLike:
        other_to_json = self.other_adapter.to_json_expr(in_expr, readable)
        if other_to_json == in_expr:
            return in_expr
        return Expr.join(
            "(None if ",
            in_expr,
            " is None else ",
            other_to_json,
            ")",
        )

    def from_json_expr(self, json_expr: ExprLike) -> ExprLike:
        other_from_json = self.other_adapter.from_json_expr(json_expr)
        if other_from_json == json_expr:
            return json_expr
        return Expr.join(
            "(None if ",
            json_expr,
            " is None else ",
            other_from_json,
            ")",
        )

    def finalize(
        self,
        resolve_type_fn: Callable[[_spec.Type], "TypeAdapter"],
    ) -> None:
        self.other_adapter.finalize(resolve_type_fn)

    def get_type(self) -> reflection.Type:
        return reflection.OptionalType(
            kind="optional",
            value=self.other_adapter.get_type(),
        )

    def register_records(
        self,
        registry: dict[str, reflection.Record],
    ) -> None:
        self.other_adapter.register_records(registry)


_other_adapter_to_optional_adapter: WeakValueDictionary[TypeAdapter, TypeAdapter] = (
    WeakValueDictionary()
)



import abc
from typing import Generic, Optional, TypeVar

Item = TypeVar("Item")
Key = TypeVar("Key")


class KeyedItems(Generic[Item, Key], tuple[Item, ...], metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def find(self, key: Key) -> Optional[Item]: ...

    @abc.abstractmethod
    def find_or_default(self, key: Key) -> Item: ...



import copy
from collections.abc import Callable, Sequence
from dataclasses import FrozenInstanceError, dataclass
from typing import Any, Final, Union

from soia._impl.function_maker import BodyBuilder, Expr, ExprLike, Line, make_function
from soia._impl.repr import repr_impl
from soia._impl.type_adapter import TypeAdapter

from soia import _spec, reflection


class EnumAdapter(TypeAdapter):
    __slots__ = (
        "spec",
        "gen_class",
        "private_is_enum_attr",
        "finalization_state",
        "value_fields",
    )

    spec: Final[_spec.Enum]
    gen_class: Final[type]  # AKA the base class
    private_is_enum_attr: Final[str]
    # 0: has not started; 1: in progress; 2: done
    finalization_state: int
    value_fields: tuple["_ValueField", ...]

    def __init__(self, spec: _spec.Enum):
        self.finalization_state = 0
        self.spec = spec
        base_class = self.gen_class = _make_base_class(spec)

        private_is_enum_attr = _name_private_is_enum_attr(spec.id)
        self.private_is_enum_attr = private_is_enum_attr
        setattr(base_class, private_is_enum_attr, True)

        # Add the constants.
        for constant_field in self.all_constant_fields:
            constant_class = _make_constant_class(base_class, constant_field)
            constant = constant_class()
            setattr(base_class, constant_field.attribute, constant)

    def finalize(
        self,
        resolve_type_fn: Callable[[_spec.Type], TypeAdapter],
    ) -> None:
        if self.finalization_state != 0:
            # Finalization is either in progress or done.
            return
        # Mark finalization as in progress.
        self.finalization_state = 1

        base_class = self.gen_class

        # Resolve the type of every value field.
        self.value_fields = value_fields = tuple(
            _make_value_field(f, resolve_type_fn(f.type), base_class)
            for f in self.spec.value_fields
        )

        # Aim to have dependencies finalized *before* the dependent. It's not always
        # possible, because there can be cyclic dependencies.
        # The function returned by the do_x_fn() method of a dependency is marginally
        # faster if the dependency is finalized. If the dependency is not finalized,
        # this function is a "forwarding" function.
        for value_field in value_fields:
            value_field.field_type.finalize(resolve_type_fn)

        # Add the wrap static factory methods.
        for value_field in value_fields:
            wrap_fn = _make_wrap_fn(value_field)
            setattr(base_class, f"wrap_{value_field.spec.name}", wrap_fn)

        base_class._fj = _make_from_json_fn(
            self.all_constant_fields,
            value_fields,
            set(self.spec.removed_numbers),
            base_class,
        )

        # Mark finalization as done.
        self.finalization_state = 2

    @property
    def all_constant_fields(self) -> list[_spec.ConstantField]:
        unknown_field = _spec.ConstantField(
            name="?",
            number=0,
            _attribute="UNKNOWN",
        )
        return list(self.spec.constant_fields) + [unknown_field]

    def default_expr(self) -> Expr:
        return Expr.local("_d?", self.gen_class.UNKNOWN)

    def to_frozen_expr(self, arg_expr: ExprLike) -> Expr:
        return Expr.join(
            "(",
            arg_expr,
            f".{self.private_is_enum_attr} and ",
            arg_expr,
            ")",
        )

    def is_not_default_expr(self, arg_expr: ExprLike, attr_expr: ExprLike) -> Expr:
        return Expr.join(arg_expr, "._number")

    def to_json_expr(self, in_expr: ExprLike, readable: bool) -> Expr:
        return Expr.join(in_expr, "._rj" if readable else "._dj")

    def from_json_expr(self, json_expr: ExprLike) -> Expr:
        fn_name = "_fj"
        from_json_fn = getattr(self.gen_class, fn_name, None)
        if from_json_fn:
            return Expr.join(Expr.local("_fj?", from_json_fn), "(", json_expr, ")")
        else:
            return Expr.join(
                Expr.local("_cls?", self.gen_class), f".{fn_name}(", json_expr, ")"
            )

    def get_type(self) -> reflection.Type:
        return reflection.RecordType(
            kind="record",
            value=self.spec.id,
        )

    def register_records(
        self,
        registry: dict[str, reflection.Record],
    ) -> None:
        record_id = self.spec.id
        if record_id in registry:
            return
        registry[record_id] = reflection.Record(
            kind="enum",
            id=record_id,
            fields=tuple(
                reflection.Field(
                    name=field.spec.name,
                    number=field.spec.number,
                    type=field.field_type.get_type(),
                )
                for field in self.value_fields
            ),
            removed_fields=self.spec.removed_numbers,
        )
        for field in self.value_fields:
            field.field_type.register_records(registry)


def _make_base_class(spec: _spec.Enum) -> type:
    record_hash = hash(spec.id)

    class BaseClass:
        __slots__ = ("value",)

        kind: str
        value: Any

        def __init__(self, never: Any):
            raise TypeError("Cannot call the constructor of a soia enum")

        @property
        def union(self) -> Any:
            return self

        def __setattr__(self, name: str, value: Any):
            raise FrozenInstanceError(self.__class__.__qualname__)

        def __delattr__(self, name: str):
            raise FrozenInstanceError(self.__class__.__qualname__)

        def __eq__(self, other: Any) -> bool:
            if isinstance(other, BaseClass):
                return other.kind == self.kind and other.value == self.value
            return NotImplemented

        def __hash__(self) -> int:
            return hash((record_hash, self.kind, self.value))

    BaseClass.__name__ = spec.class_name
    BaseClass.__qualname__ = spec.class_qualname

    return BaseClass


def _make_constant_class(base_class: type, spec: _spec.ConstantField) -> type:
    class Constant(base_class):
        __slots__ = ()

        kind: Final[str] = spec.name
        _number: Final[int] = spec.number
        # dense JSON
        _dj: Final[int] = spec.number
        # readable JSON
        _rj: Final[str] = spec.name
        # has value
        _hv: Final[bool] = False

        def __init__(self):
            # Do not call super().__init__().
            object.__setattr__(self, "value", None)

        def __repr__(self) -> str:
            return f"{base_class.__qualname__}.{spec.attribute}"

    return Constant


def _make_unrecognized_class(base_class: type) -> type:
    """Wraps around an unrecognized dense JSON.

    Looks and acts just like the UNKNOWN constant, except that its JSON representation
    is the original unrecognized dense JSON.
    """

    class Unrecognized(base_class):
        __slots__ = ("_dj",)

        kind: Final[str] = "?"
        _number: Final[int] = 0
        # dense JSON
        _dj: Any
        # readable JSON
        _rj: Final[str] = "?"
        # has value
        _hv: Final[bool] = False

        def __init__(self, dj: Any):
            # Do not call super().__init__().
            object.__setattr__(self, "_dj", copy.deepcopy(dj))
            object.__setattr__(self, "value", None)

        def __repr__(self) -> str:
            return f"{base_class.__qualname__}.UNKNOWN"

    return Unrecognized


def _make_value_class(
    base_class: type,
    field_spec: _spec.ValueField,
    field_type: TypeAdapter,
) -> type:
    number = field_spec.number

    class Value(base_class):
        __slots__ = ()

        kind: Final[str] = field_spec.name
        _number: Final[int] = number
        # has value
        _hv: Final[bool] = True

        def __init__(self):
            # Do not call super().__init__().
            pass

        def __repr__(self) -> str:
            value_repr = repr_impl(self.value)
            if value_repr.complex:
                body = f"\n  {value_repr.indented}\n"
            else:
                body = value_repr.repr
            return f"{base_class.__qualname__}.wrap_{field_spec.name}({body})"

    ret = Value

    ret._dj = property(
        make_function(
            name="to_dense_json",
            params=["self"],
            body=[
                Line.join(
                    f"return [{field_spec.number}, ",
                    field_type.to_json_expr("self.value", readable=False),
                    "]",
                ),
            ],
        )
    )

    ret._rj = property(
        make_function(
            name="to_readable_json",
            params=["self"],
            body=[
                Line.join(
                    "return {",
                    f'"kind": "{field_spec.name}", "value": ',
                    field_type.to_json_expr("self.value", readable=True),
                    "}",
                ),
            ],
        )
    )

    return ret


@dataclass(frozen=True)
class _ValueField:
    spec: _spec.ValueField
    field_type: TypeAdapter
    value_class: type


def _make_value_field(
    spec: _spec.ValueField, field_type: TypeAdapter, base_class: type
) -> _ValueField:
    return _ValueField(
        spec=spec,
        field_type=field_type,
        value_class=_make_value_class(
            base_class=base_class, field_spec=spec, field_type=field_type
        ),
    )


def _make_wrap_fn(field: _ValueField) -> Callable[[Any], Any]:
    builder = BodyBuilder()
    builder.append_ln("ret = ", Expr.local("value_class", field.value_class), "()")
    builder.append_ln(
        Expr.local("setattr", object.__setattr__),
        "(ret, 'value', ",
        field.field_type.to_frozen_expr("value"),
        ")",
    )
    builder.append_ln("return ret")
    return make_function(
        name="wrap",
        params=["value"],
        body=builder.build(),
    )


def _make_from_json_fn(
    constant_fields: Sequence[_spec.ConstantField],
    value_fields: Sequence[_ValueField],
    removed_numbers: set[int],
    base_class: type,
) -> Callable[[Any], Any]:
    unrecognized_class = _make_unrecognized_class(base_class)
    unrecognized_class_local = Expr.local("Unrecognized", unrecognized_class)
    obj_setattr_local = Expr.local("obj_settatr", object.__setattr__)
    removed_numbers_local = Expr.local("removed_numbers", removed_numbers)

    key_to_constant: dict[Union[int, str], Any] = {}
    for field in constant_fields:
        constant = getattr(base_class, field.attribute)
        key_to_constant[field.number] = constant
        key_to_constant[field.name] = constant
    key_to_constant_local = Expr.local("key_to_constant", key_to_constant)
    unknown_constant = key_to_constant[0]
    unknown_constant_local = Expr.local("unknown_constant", unknown_constant)

    numbers: list[int] = []
    names: list[str] = []
    key_to_field: dict[Union[int, str], _ValueField] = {}
    for field in value_fields:
        numbers.append(field.spec.number)
        names.append(field.spec.name)
        key_to_field[field.spec.number] = field
        key_to_field[field.spec.name] = field
    value_keys_local = Expr.local("value_keys", set(key_to_field.keys()))

    builder = BodyBuilder()
    # The reason why we wrap the function inside a 'while' is explained below.
    builder.append_ln("while True:")

    # DENSE FORMAT
    if len(constant_fields) == 1:
        builder.append_ln("  if json == 0:")
        builder.append_ln("    return ", unknown_constant_local)
    else:
        # `json.__class__ is int` is significantly faster than `isinstance(json, int)`
        builder.append_ln("  if json.__class__ is int:")
        builder.append_ln("    try:")
        builder.append_ln("      return ", key_to_constant_local, "[json]")
        builder.append_ln("    except:")
        if removed_numbers:
            builder.append_ln("      if json in ", removed_numbers_local, ":")
            builder.append_ln("        return ", unknown_constant_local)
        builder.append_ln("      return ", unrecognized_class_local, "(json)")

    def append_number_branches(numbers: list[int], indent: str) -> None:
        if len(numbers) == 1:
            number = numbers[0]
            field = key_to_field[number]
            value_class_local = Expr.local("cls?", field.value_class)
            value_expr = field.field_type.from_json_expr("json[1]")
            builder.append_ln(f"{indent}ret = ", value_class_local, "()")
            builder.append_ln(
                indent, obj_setattr_local, '(ret, "value", ', value_expr, ")"
            )
            builder.append_ln(f"{indent}return ret")
        else:
            indented = f"  {indent}"
            mid_index = int(len(numbers) / 2)
            mid_number = numbers[mid_index - 1]
            operator = "==" if mid_index == 1 else "<="
            builder.append_ln(f"{indent}if number {operator} {mid_number}:")
            append_number_branches(numbers[0:mid_index], indented)
            builder.append_ln(f"{indent}else:")
            append_number_branches(numbers[mid_index:], indented)

    # `json.__class__ is list` is significantly faster than `isinstance(json, list)`
    builder.append_ln("  elif json.__class__ is list:")
    builder.append_ln("    number = json[0]")
    if not value_fields:
        # The field was either removed or is an unrecognized field.
        if removed_numbers:
            builder.append_ln("    if number in ", removed_numbers_local, ":")
            builder.append_ln("      return ", unknown_constant_local)
        builder.append_ln("    return ", unrecognized_class_local, "(json)")
    else:
        if len(value_fields) == 1:
            builder.append_ln(f"    if number != {value_fields[0].spec.number}:")
        else:
            builder.append_ln("    if number not in ", value_keys_local, ":")
        if removed_numbers:
            builder.append_ln("      if number in ", removed_numbers_local, ":")
            builder.append_ln("        return ", unknown_constant_local)
        builder.append_ln("      return ", unrecognized_class_local, "(json)")
        append_number_branches(sorted(numbers), "    ")

    # READABLE FORMAT
    if len(constant_fields) == 1:
        builder.append_ln("  elif json == '?':")
        builder.append_ln("    return ", unknown_constant_local)
    else:
        builder.append_ln("  if isinstance(json, str):")
        builder.append_ln("    try:")
        builder.append_ln("      return ", key_to_constant_local, "[json]")
        builder.append_ln("    except:")
        # In readable mode, drop unrecognized values and use UNKNOWN instead.
        builder.append_ln("      return ", unknown_constant_local)

    def append_name_branches(names: list[str], indent: str) -> None:
        if len(names) == 1:
            name = names[0]
            field = key_to_field[name]
            value_class_local = Expr.local("cls?", field.value_class)
            value_expr = field.field_type.from_json_expr("json['value']")
            builder.append_ln(f"{indent}ret = ", value_class_local, "()")
            builder.append_ln(
                indent, obj_setattr_local, '(ret, "value", ', value_expr, ")"
            )
            builder.append_ln(f"{indent}return ret")
        else:
            indented = f"  {indent}"
            mid_index = int(len(names) / 2)
            mid_name = names[mid_index - 1]
            operator = "==" if mid_index == 1 else "<="
            builder.append_ln(f"{indent}if kind {operator} '{mid_name}':")
            append_name_branches(names[0:mid_index], indented)
            builder.append_ln(f"{indent}else:")
            append_name_branches(names[mid_index:], indented)

    builder.append_ln("  elif isinstance(json, dict):")
    if not value_fields:
        builder.append_ln("    return ", unknown_constant_local)
    else:
        builder.append_ln("    kind = json['kind']")
        builder.append_ln("    if kind not in ", value_keys_local, ":")
        builder.append_ln("      return ", unknown_constant_local)
        builder.append_ln("    else:")
        append_name_branches(sorted(names), "      ")

    # In the unlikely event that json.loads() returns an instance of a subclass of int.
    builder.append_ln("  elif isinstance(json, int):")
    builder.append_ln("    json = int(json)")
    builder.append_ln("  elif isinstance(json, list):")
    builder.append_ln("    json = list(json)")
    builder.append_ln("  else:")
    builder.append_ln("    return TypeError()")

    return make_function(
        name="from_json",
        params=["json"],
        body=builder.build(),
    )


def _name_private_is_enum_attr(record_id: str) -> str:
    record_name = _spec.RecordId.parse(record_id).name
    hex_hash = hex(abs(hash(record_id)))[:6]
    return f"_is_{record_name}_{hex_hash}"



import copy
from collections.abc import Callable, Sequence
from dataclasses import FrozenInstanceError, dataclass
from typing import Any, Final, Union, cast

from soia._impl.function_maker import (
    BodyBuilder,
    Expr,
    ExprLike,
    Line,
    LineSpan,
    LineSpanLike,
    Param,
    Params,
    make_function,
)
from soia._impl.repr import repr_impl
from soia._impl.type_adapter import TypeAdapter

from soia import _spec, reflection


class StructAdapter(TypeAdapter):
    __slots__ = (
        "spec",
        "record_hash",
        "gen_class",
        "mutable_class",
        "simple_class",
        "private_is_frozen_attr",
        "finalization_state",
        "fields",
    )

    spec: Final[_spec.Struct]
    record_hash: Final[int]
    # The frozen class.
    gen_class: Final[type]
    mutable_class: Final[type]
    simple_class: Final[type]
    private_is_frozen_attr: Final[str]

    # 0: has not started; 1: in progress; 2: done
    finalization_state: int
    fields: tuple["_Field", ...]

    def __init__(self, spec: _spec.Struct):
        self.finalization_state = 0
        self.spec = spec
        self.record_hash = hash(spec.id)

        self.private_is_frozen_attr = _name_private_is_frozen_attr(spec.id)

        slots = tuple(f.attribute for f in self.spec.fields) + (
            # Unrecognized fields encountered during deserialization.
            "_unrecognized",
            # Lowest number greater than the number of every field with a non-default
            # value.
            "_array_len",
        )
        frozen_class = self.gen_class = _make_dataclass(slots)
        frozen_class.__name__ = spec.class_name
        frozen_class.__qualname__ = spec.class_qualname
        frozen_class.__setattr__ = cast(Any, _Frozen.__setattr__)
        frozen_class.__delattr__ = cast(Any, _Frozen.__delattr__)
        # We haven't added an __init__ method to the frozen class yet, so frozen_class()
        # returns an object with no attribute set. We'll set the attributes of DEFAULT
        # at the finalization step.
        frozen_class.DEFAULT = frozen_class()

        mutable_class = self.mutable_class = _make_dataclass(slots)
        mutable_class.__name__ = "Mutable"
        mutable_class.__qualname__ = f"{spec.class_qualname}.Mutable"
        frozen_class.Mutable = mutable_class
        frozen_class.OrMutable = Union[frozen_class, mutable_class]

        # The 'simple' class is a class which only has __slots__ defined.
        # To construct instances of the frozen class from internal implementation,
        # instead of calling the constructor of the frozen class, we use the following
        # technique: create an instance of the simple class, assign all the attributes,
        # then assign the frozen class to the __class__ attribute.
        # Reason why it's faster: we don't need to call object.__setattr__().
        self.simple_class = _make_dataclass(slots)

    def finalize(
        self,
        resolve_type_fn: Callable[[_spec.Type], TypeAdapter],
    ) -> None:
        if self.finalization_state != 0:
            # Finalization is either in progress or done.
            return
        # Mark finalization as in progress.
        self.finalization_state = 1

        # Resolve the type of every field.
        self.fields = fields = tuple(
            sorted(
                (_Field(f, resolve_type_fn(f.type)) for f in self.spec.fields),
                key=lambda f: f.field.number,
            )
        )

        # Aim to have dependencies finalized *before* the dependent. It's not always
        # possible, because there can be cyclic dependencies.
        # The function returned by the do_x_fn() method of a dependency is marginally
        # faster if the dependency is finalized. If the dependency is not finalized,
        # this function is a "forwarding" function.
        for field in fields:
            field.type.finalize(resolve_type_fn)

        frozen_class = self.gen_class
        mutable_class = self.mutable_class
        simple_class = self.simple_class

        frozen_class.to_frozen = _identity
        frozen_class.to_mutable = _make_to_mutable_fn(
            mutable_class=mutable_class,
            simple_class=simple_class,
            fields=fields,
        )
        mutable_class.to_frozen = _make_to_frozen_fn(
            frozen_class=frozen_class,
            simple_class=simple_class,
            fields=fields,
        )
        setattr(frozen_class, self.private_is_frozen_attr, True)
        setattr(mutable_class, self.private_is_frozen_attr, False)

        frozen_class.__init__ = cast(
            Any,
            _make_frozen_class_init_fn(
                fields,
                frozen_class=frozen_class,
                simple_class=simple_class,
            ),
        )
        mutable_class.__init__ = cast(Any, _make_mutable_class_init_fn(fields))
        frozen_class.whole = _make_whole_static_factory_method(frozen_class)

        frozen_class.__eq__ = _make_eq_fn(fields)
        frozen_class.__hash__ = cast(Any, _make_hash_fn(fields, self.record_hash))
        frozen_class.__repr__ = mutable_class.__repr__ = cast(
            Any, _make_repr_fn(fields)
        )

        frozen_class._tdj = _make_to_dense_json_fn(fields=fields)
        frozen_class._trj = _make_to_readable_json_fn(fields=fields)
        frozen_class._fj = _make_from_json_fn(
            frozen_class=frozen_class,
            simple_class=simple_class,
            fields=fields,
            removed_numbers=self.spec.removed_numbers,
        )

        # Initialize DEFAULT.
        _init_default(self.gen_class.DEFAULT, fields)

        # Define mutable getters
        for field in fields:
            if field.field.has_mutable_getter:
                mutable_getter = _make_mutable_getter(field)
                mutable_getter_name = mutable_getter.__name__
                setattr(mutable_class, mutable_getter_name, property(mutable_getter))

        # Mark finalization as done.
        self.finalization_state = 2

    def default_expr(self) -> Expr:
        return Expr.local("_d?", self.gen_class.DEFAULT)

    def to_frozen_expr(self, arg_expr: ExprLike) -> Expr:
        # The goal of referring to private_is_frozen_attr is to raise an error if the
        # struct has the wrong type.
        return Expr.join(
            "(",
            arg_expr,
            " if ",
            arg_expr,
            f".{self.private_is_frozen_attr} else ",
            arg_expr,
            ".to_frozen())",
        )

    def is_not_default_expr(self, arg_expr: ExprLike, attr_expr: ExprLike) -> Expr:
        return Expr.join(attr_expr, "._array_len")

    def to_json_expr(
        self,
        in_expr: ExprLike,
        readable: bool,
    ) -> Expr:
        return Expr.join(in_expr, "._trj" if readable else "._tdj", "()")

    def from_json_expr(self, json_expr: ExprLike) -> Expr:
        fn_name = "_fj"
        # The _fj method may not have been added to the class yet.
        from_json_fn = getattr(self.gen_class, fn_name, None)
        if from_json_fn:
            return Expr.join(Expr.local("_fj?", from_json_fn), "(", json_expr, ")")
        else:
            return Expr.join(
                Expr.local("_cls?", self.gen_class), f".{fn_name}(", json_expr, ")"
            )

    def get_type(self) -> reflection.Type:
        return reflection.RecordType(
            kind="record",
            value=self.spec.id,
        )

    def register_records(
        self,
        registry: dict[str, reflection.Record],
    ) -> None:
        record_id = self.spec.id
        if record_id in registry:
            return
        registry[record_id] = reflection.Record(
            kind="struct",
            id=record_id,
            fields=tuple(
                reflection.Field(
                    name=field.field.name,
                    number=field.field.number,
                    type=field.type.get_type(),
                )
                for field in self.fields
            ),
            removed_fields=self.spec.removed_numbers,
        )
        for field in self.fields:
            field.type.register_records(registry)


class _Frozen:
    def __setattr__(self, name: str, value: Any):
        raise FrozenInstanceError(self.__class__.__qualname__)

    def __delattr__(self, name: str):
        raise FrozenInstanceError(self.__class__.__qualname__)


def _make_dataclass(slots: tuple[str, ...]) -> type:
    class Result:
        __slots__ = slots

    return Result


def _identity(input: Any) -> Any:
    return input


@dataclass(frozen=True)
class _Field:
    field: _spec.Field
    type: TypeAdapter


def _make_frozen_class_init_fn(
    fields: Sequence[_Field],
    frozen_class: type,
    simple_class: type,
) -> Callable[..., None]:
    """
    Returns the implementation of the __init__() method of the frozen class.
    """

    # Set the params.
    params: Params = ["_self"]
    if fields:
        params.append("*")
    for field in fields:
        params.append(
            Param(
                name=field.field.attribute,
                default=field.type.default_expr(),
            )
        )

    builder = BodyBuilder()
    # Since __setattr__() was overridden to raise errors in order to make the class
    # immutable, the only way to set attributes in the constructor is to call
    # object.__setattr__().
    obj_setattr = Expr.local("_setattr", object.__setattr__)

    def array_len_expr() -> Expr:
        spans: list[LineSpanLike] = []
        # Fields are sorted by number.
        for field in reversed(fields):
            spans.append(
                LineSpan.join(
                    f"{field.field.number + 1} if ",
                    field.type.is_not_default_expr(
                        arg_expr=field.field.attribute,
                        attr_expr=f"_self.{field.field.attribute}",
                    ),
                    " else ",
                )
            )
        spans.append("0")
        return Expr.join(*spans)

    if len(fields) < 3:
        # If the class has less than 3 fields, it is faster to assign every field with
        # object.__setattr__().
        for field in fields:
            attribute = field.field.attribute
            builder.append_ln(
                obj_setattr,
                '(_self, "',
                attribute,
                '", ',
                field.type.to_frozen_expr(attribute),
                ")",
            )
        # Set the _unrecognized field.
        builder.append_ln(obj_setattr, '(_self, "_unrecognized", ())')
        # Set array length.
        builder.append_ln(
            obj_setattr,
            '(_self, "_array_len", ',
            array_len_expr(),
            ")",
        )
    else:
        # If the class has 4 fields or more, it is faster to first change the __class__
        # attribute of the object so it's no longer immutable, assign all the attributes
        # with regular assignment, and then change back the __class__ attribute.
        builder.append_ln(
            obj_setattr,
            '(_self, "__class__", ',
            Expr.local("Simple", simple_class),
            ")",
        )
        for field in fields:
            attribute = field.field.attribute
            builder.append_ln(
                "_self.",
                attribute,
                " = ",
                field.type.to_frozen_expr(attribute),
            )
        # Set the _unrecognized field.
        builder.append_ln("_self._unrecognized = ()")
        # Set array length.
        builder.append_ln("_self._array_len = ", array_len_expr())
        # Change back the __class__.
        builder.append_ln("_self.__class__ = ", Expr.local("Frozen", frozen_class))

    return make_function(
        name="__init__",
        params=params,
        body=builder.build(),
    )


def _make_mutable_class_init_fn(fields: Sequence[_Field]) -> Callable[..., None]:
    """
    Returns the implementation of the __init__() method of the mutable class.
    """

    params: Params = ["_self"]
    builder = BodyBuilder()
    for field in fields:
        attribute = field.field.attribute
        params.append(
            Param(
                name=attribute,
                default=field.type.default_expr(),
            )
        )
        builder.append_ln(
            "_self.",
            attribute,
            " = ",
            attribute,
        )
    builder.append_ln("_self._unrecognized = ()")
    return make_function(
        name="__init__",
        params=params,
        body=builder.build(),
    )


def _make_whole_static_factory_method(frozen_class: type) -> Callable[..., Any]:
    def whole(**kwargs):
        return frozen_class(**kwargs)
    return whole


def _make_to_mutable_fn(
    mutable_class: type,
    simple_class: type,
    fields: Sequence[_Field],
) -> Callable[[Any], Any]:
    """
    Returns the implementation of the to_mutable() method of the frozen class.
    """
    builder = BodyBuilder()
    # Create an instance of the simple class. We'll later change its __class__ attr.
    builder.append_ln(
        "ret = ",
        Expr.local("Simple", simple_class),
        "()",
    )
    for field in fields:
        attribute = field.field.attribute
        builder.append_ln("ret.", attribute, " = self.", attribute)
    builder.append_ln("ret._unrecognized = self._unrecognized")
    builder.append_ln(
        "ret.__class__ = ",
        Expr.local("mutable_class", mutable_class),
    )
    builder.append_ln("return ret")
    return make_function(
        name="to_mutable",
        params=["self"],
        body=builder.build(),
    )


def _make_to_frozen_fn(
    frozen_class: type,
    simple_class: type,
    fields: Sequence[_Field],
) -> Callable[[Any], Any]:
    """
    Returns the implementation of the to_frozen() method of the mutable class.
    """

    builder = BodyBuilder()
    # Create an instance of the simple class. We'll later change its __class__ attr.
    builder.append_ln(
        "ret = ",
        Expr.local("Simple", simple_class),
        "()",
    )
    for field in fields:
        attribute = field.field.attribute
        builder.append_ln(
            "ret.",
            attribute,
            " = ",
            field.type.to_frozen_expr(f"self.{attribute}"),
        )

    # Set the _unrecognized field.
    builder.append_ln("ret._unrecognized = self._unrecognized")

    def array_len_expr() -> Expr:
        spans: list[LineSpanLike] = []
        spans.append(
            LineSpan.join(
                f"{_get_num_slots(fields)} + ",
                Expr.local("_len", len),
                "(ret._unrecognized) if ret._unrecognized else ",
            )
        )
        # Fields are sorted by number.
        for field in reversed(fields):
            attr_expr = f"ret.{field.field.attribute}"
            spans.append(
                LineSpan.join(
                    f"{field.field.number + 1} if ",
                    field.type.is_not_default_expr(attr_expr, attr_expr),
                    " else ",
                )
            )
        spans.append("0")
        return Expr.join(*spans)

    # Set the _unrecognized field.
    builder.append_ln("ret._unrecognized = self._unrecognized")
    # Set array length.
    builder.append_ln("ret._array_len = ", array_len_expr())

    builder.append_ln(
        "ret.__class__ = ",
        Expr.local("Frozen", frozen_class),
    )
    builder.append_ln("return ret")
    return make_function(
        name="to_frozen",
        params=["self"],
        body=builder.build(),
    )


def _make_eq_fn(
    fields: Sequence[_Field],
) -> Callable[[Any], Any]:
    """
    Returns the implementation of the __eq__() method of the frozen class.
    """

    builder = BodyBuilder()
    builder.append_ln("if other is self:")
    builder.append_ln("  return True")
    builder.append_ln("if other.__class__ is self.__class__:")
    operands: list[ExprLike]
    if fields:
        attr: Callable[[_Field], str] = lambda f: f.field.attribute
        operands = [Expr.join(f"self.{attr(f)} == other.{attr(f)}") for f in fields]
    else:
        operands = ["True"]
    builder.append_ln("  return ", Expr.join(*operands, separator=" and "))
    builder.append_ln("return ", Expr.local("NotImplemented", NotImplemented))
    return make_function(
        name="__eq__",
        params=["self", "other"],
        body=builder.build(),
    )


def _make_hash_fn(
    fields: Sequence[_Field],
    record_hash: int,
) -> Callable[[Any], Any]:
    """
    Returns the implementation of the __hash__() method of the frozen class.
    """

    components: list[ExprLike] = []
    components.append(str(record_hash))
    for field in fields:
        components.append(f"self.{field.field.attribute}")
    return make_function(
        name="__hash__",
        params=["self"],
        body=[Line.join("return hash((", Expr.join(*components, separator=", "), "))")],
    )


def _make_repr_fn(fields: Sequence[_Field]) -> Callable[[Any], str]:
    """
    Returns the implementation of the __repr__() method of both the frozen class and the
    mutable class.
    """

    builder = BodyBuilder()
    builder.append_ln("if self is getattr(self.__class__, 'DEFAULT', None):")
    builder.append_ln("  return f'{self.__class__.__qualname__}.DEFAULT'")
    builder.append_ln("assignments = []")
    builder.append_ln("any_complex = False")
    builder.append_ln("is_mutable = self.__class__.__name__ == 'Mutable'")
    repr_local = Expr.local("repr", repr_impl)
    for field in fields:
        attribute = field.field.attribute
        # is_not_default_expr only works on a frozen expression.
        to_frozen_expr = field.type.to_frozen_expr(f"(self.{attribute})")
        is_not_default_expr = field.type.is_not_default_expr(
            to_frozen_expr, to_frozen_expr
        )
        builder.append_ln("if is_mutable or ", is_not_default_expr, ":")
        builder.append_ln("  r = ", repr_local, f"(self.{attribute})")
        builder.append_ln(f"  assignments.append(f'{attribute}={{r.indented}}')")
        builder.append_ln("  any_complex = any_complex or r.complex")
    builder.append_ln("if len(assignments) <= 1 and not any_complex:")
    builder.append_ln("  body = ''.join(assignments)")
    builder.append_ln("else:")
    builder.append_ln("  body = '\\n' + ''.join(f'  {a},\\n' for a in assignments)")
    builder.append_ln("return f'{self.__class__.__qualname__}({body})'")

    return make_function(
        name="__repr__",
        params=["self"],
        body=builder.build(),
    )


def _make_to_dense_json_fn(fields: Sequence[_Field]) -> Callable[[Any], Any]:
    builder = BodyBuilder()
    builder.append_ln("l = self._array_len")
    builder.append_ln("ret = [0] * l")
    for field in fields:
        builder.append_ln(f"if l <= {field.field.number}:")
        builder.append_ln("  return ret")
        builder.append_ln(
            f"ret[{field.field.number}] = ",
            field.type.to_json_expr(
                f"self.{field.field.attribute}",
                readable=False,
            ),
        )
    num_slots = _get_num_slots(fields)
    builder.append_ln(f"if l <= {num_slots}:")
    builder.append_ln("  return ret")
    builder.append_ln(f"ret[{num_slots}:] = self._unrecognized")
    builder.append_ln("return ret")
    return make_function(
        name="to_dense_json",
        params=["self"],
        body=builder.build(),
    )


def _make_to_readable_json_fn(fields: Sequence[_Field]) -> Callable[[Any], Any]:
    builder = BodyBuilder()
    builder.append_ln("ret = {}")
    for field in fields:
        attr_expr = f"self.{field.field.attribute}"
        builder.append_ln(
            "if ",
            field.type.is_not_default_expr(attr_expr, attr_expr),
            ":",
        )
        builder.append_ln(
            f'  ret["{field.field.name}"] = ',
            field.type.to_json_expr(
                attr_expr,
                readable=True,
            ),
        )
    builder.append_ln("return ret")
    return make_function(
        name="to_readable_json",
        params=["self"],
        body=builder.build(),
    )


def _make_from_json_fn(
    frozen_class: type,
    simple_class: type,
    fields: Sequence[_Field],
    removed_numbers: tuple[int, ...],
) -> Callable[[Any], Any]:
    builder = BodyBuilder()
    builder.append_ln("if not json:")
    builder.append_ln("  return ", Expr.local("DEFAULT", frozen_class.DEFAULT))
    builder.append_ln("ret = ", Expr.local("Simple", simple_class), "()")
    builder.append_ln(
        "if ",
        Expr.local("isinstance", isinstance),
        "(json, ",
        Expr.local("list", list),
        "):",
    )
    # JSON array (dense flavor)
    builder.append_ln(
        "  array_len = ",
        Expr.local("len", len),
        "(json)",
    )
    for field in fields:
        name = field.field.name
        number = field.field.number
        item_expr = f"json[{number}]"
        builder.append_ln(
            f"  ret.{field.field.attribute} = ",
            field.type.default_expr(),
            f" if array_len <= {number} else ",
            field.type.from_json_expr(item_expr),
        )
    num_slots = _get_num_slots(fields)
    builder.append_ln(f"  if array_len <= {num_slots}:")
    builder.append_ln("    ret._unrecognized = ()")
    builder.append_ln(
        "    ret._array_len = ",
        _adjust_array_len_expr("array_len", removed_numbers),
    )
    builder.append_ln("  else:")
    builder.append_ln(
        "    ret._unrecognized = ",
        Expr.local("deepcopy", copy.deepcopy),
        f"(json[{num_slots}:])",
    )
    builder.append_ln("    ret._array_len = array_len")

    builder.append_ln("else:")
    builder.append_ln("  array_len = 0")
    # JSON object (readable flavor)
    for field in fields:
        name = field.field.name
        lvalue = f"ret.{field.field.attribute}"
        builder.append_ln(f'  if "{name}" in json:')
        builder.append_ln(f"    array_len = {field.field.number + 1}")
        builder.append_ln(
            f"    {lvalue} = ",
            field.type.from_json_expr(f'json["{name}"]'),
        )
        builder.append_ln("  else:")
        builder.append_ln(f"    {lvalue} = ", field.type.default_expr())
    # Drop unrecognized fields in readable mode.
    builder.append_ln("  ret._unrecognized = ()")
    builder.append_ln("  ret._array_len = array_len")

    builder.append_ln("ret.__class__ = ", Expr.local("Frozen", frozen_class))
    builder.append_ln("return ret")

    return make_function(
        name="from_json",
        params=["json"],
        body=builder.build(),
    )


def _adjust_array_len_expr(var: str, removed_numbers: tuple[int, ...]) -> str:
    """
    When parsing a dense JSON or decoding a binary string, we can reuse the array length
    in the decoded struct, but we need to account for possibly newly-removed fields. The
    last field of the adjusted array length cannot be a removed field.

    Let's imagine that field number 3 was removed from a struct.
    This function would return the following expression:
        array_len if array_len <= 3 else 3 if array_len == 4 else array_len
    """

    @dataclass
    class _RemovedSpan:
        """Sequence of consecutive removed fields."""

        # Number of the first removed field.
        begin: int = 0
        # Number after the last removed field.
        end: int = 0

    def get_removed_spans() -> list[_RemovedSpan]:
        ret: list[_RemovedSpan] = []
        for number in sorted(removed_numbers):
            last = ret[-1] if ret else None
            if last and last.end == number:
                last.end = number + 1
            else:
                ret.append(_RemovedSpan(number, number + 1))
        return ret

    removed_spans = get_removed_spans()

    ret = ""
    lower_bound = 0
    for s in removed_spans:
        if s.begin == lower_bound:
            ret += f"{s.begin} if {var} <= {s.end} else "
        elif s.end == s.begin + 1:
            # Similar to the expression in 'else' but uses '==' instead of '<='
            ret += (
                f"{var} if {var} <= {s.begin} else {s.begin} if {var} == {s.end} else "
            )
        else:
            ret += (
                f"{var} if {var} <= {s.begin} else {s.begin} if {var} <= {s.end} else "
            )
        lower_bound = s.end + 1
    ret += var
    return ret


def _init_default(default: Any, fields: Sequence[_Field]) -> None:
    for field in fields:
        attribute = field.field.attribute
        get_field_default = make_function(
            name="get_default",
            params=(),
            body=(Line.join("return ", field.type.default_expr()),),
        )
        object.__setattr__(default, attribute, get_field_default())
    object.__setattr__(default, "_unrecognized", ())
    object.__setattr__(default, "_array_len", 0)


def _make_mutable_getter(field: _Field) -> Callable[[Any], Any]:
    # Two cases: the field either has struct type or array type.
    attribute = field.field.attribute
    builder = BodyBuilder()
    if isinstance(field.type, StructAdapter):
        frozen_class = field.type.gen_class
        mutable_class = frozen_class.Mutable
        builder.append_ln(
            f"if self.{attribute}.__class__ is ",
            Expr.local("mutable_class", mutable_class),
            ":",
        )
        builder.append_ln(f"  return self.{attribute}")
        builder.append_ln(
            f"if self.{attribute}.__class__ is ",
            Expr.local("frozen_class", frozen_class),
            ":",
        )
        builder.append_ln(f"  self.{attribute} = self.{attribute}.to_mutable()")
        builder.append_ln(f"  return self.{attribute}")
        expected = f"{frozen_class.__qualname__} or {mutable_class.__qualname__}"
        found = f"self.{attribute}.__class__.__name__"
        builder.append_ln(
            "raise ",
            Expr.local("TypeError", TypeError),
            f"(f'expected: {expected}; found: {{{found}}}')",
        )
    else:
        builder.append_ln(f"if not isinstance(self.{attribute}, list):")
        builder.append_ln(f"  self.{attribute} = list(self.{attribute})")
        builder.append_ln(f"return self.{attribute}")
    return make_function(
        name=f"mutable_{field.field.name}",
        params=["self"],
        body=builder.build(),
    )


def _name_private_is_frozen_attr(record_id: str) -> str:
    record_name = _spec.RecordId.parse(record_id).name
    hex_hash = hex(abs(hash(record_id)))[:6]
    return f"_is_{record_name}_{hex_hash}"


def _get_num_slots(fields: Sequence[_Field]) -> int:
    return (fields[-1].field.number + 1) if fields else 0



import json as jsonlib
from collections.abc import Callable
from dataclasses import FrozenInstanceError
from functools import cached_property
from typing import Any, Generic, TypeVar, cast, final
from weakref import WeakValueDictionary

from soia._impl.function_maker import Expr, LineSpan, make_function
from soia._impl.never import Never
from soia._impl.type_adapter import TypeAdapter

from soia import reflection

T = TypeVar("T")


@final
class Serializer(Generic[T]):
    __slots__ = (
        "__weakref__",
        "_adapter",
        "_to_dense_json_fn",
        "_to_readable_json_fn",
        "_from_json_fn",
        "__dict__",
    )

    _adapter: TypeAdapter
    _to_dense_json_fn: Callable[[T], Any]
    _to_readable_json_fn: Callable[[T], Any]
    _from_json_fn: Callable[[Any], T]

    def __init__(self, adapter: Never):
        # Use Never (^) as a trick to make the constructor internal.
        object.__setattr__(self, "_adapter", adapter)
        object.__setattr__(
            self, "_to_dense_json_fn", _make_to_json_fn(adapter, readable=False)
        )
        object.__setattr__(
            self,
            "_to_readable_json_fn",
            _make_to_json_fn(adapter, readable=True),
        )
        object.__setattr__(self, "_from_json_fn", _make_from_json_fn(adapter))

    def to_json(self, input: T, *, readable=False) -> Any:
        if readable:
            return self._to_readable_json_fn(input)
        else:
            return self._to_dense_json_fn(input)

    def to_json_code(self, input: T, readable=False) -> str:
        if readable:
            return jsonlib.dumps(self._to_readable_json_fn(input), indent=2)
        else:
            return jsonlib.dumps(self._to_dense_json_fn(input), separators=(",", ":"))

    def from_json(self, json: Any) -> T:
        return self._from_json_fn(json)

    def from_json_code(self, json_code: str) -> T:
        return self._from_json_fn(jsonlib.loads(json_code))

    @cached_property
    def type_descriptor(self) -> reflection.TypeDescriptor:
        records: dict[str, reflection.Record] = {}
        self._adapter.register_records(records)
        return reflection.TypeDescriptor(
            type=self._adapter.get_type(),
            records=tuple(records.values()),
        )

    def __setattr__(self, name: str, value: Any):
        raise FrozenInstanceError(self.__class__.__qualname__)

    def __delattr__(self, name: str):
        raise FrozenInstanceError(self.__class__.__qualname__)


# A cache to make sure we only create one Serializer for each TypeAdapter.
_type_adapter_to_serializer: WeakValueDictionary[TypeAdapter, Serializer] = (
    WeakValueDictionary()
)


def make_serializer(adapter: TypeAdapter) -> Serializer:
    return _type_adapter_to_serializer.setdefault(
        adapter, Serializer(cast(Never, adapter))
    )


def _make_to_json_fn(adapter: TypeAdapter, readable: bool) -> Callable[[Any], Any]:
    return make_function(
        name="to_json",
        params=["input"],
        body=[
            LineSpan.join(
                "return ",
                adapter.to_json_expr(
                    adapter.to_frozen_expr(Expr.join("input")),
                    readable=readable,
                ),
            ),
        ],
    )


def _make_from_json_fn(adapter: TypeAdapter) -> Callable[[Any], Any]:
    return make_function(
        name="from_json",
        params=["json"],
        body=[
            LineSpan.join("return ", adapter.from_json_expr(Expr.join("json"))),
        ],
    )
